"""
Constants used in data reduction.
"""

hdm = 6.626176e-34/1.674928e-27  # h / m
lamdaCut = 2.5  # Aa
lamdaMax = 15.0 # Aa

polarizationConfigs = ['unpolarized', 'unpolarized', 'po', 'mo', 'op', 'pp', 'mp', 'om', 'pm', 'mm']
polarizationLabels = ['undetermined', 'unpolarized', 'spin-up', 'spin-down', 'op',
                      'up-up', 'down-up', 'om', 'up-down', 'down-down']



"""
Classes for stroing various configurations needed for reduction.
"""
import argparse
from dataclasses import dataclass, field, Field, fields, MISSING
from typing import get_args, get_origin, List, Optional, Tuple, Union
from datetime import datetime
from os import path
import numpy as np

import logging


try:
    from enum import StrEnum
except ImportError:
    try:
        # python <3.11 try to use backports
        from backports.strenum import StrEnum
    except ImportError:
        # python <3.10 use Enum instead
        from enum import Enum as StrEnum

class InCallString(StrEnum):
    auto='auto'
    always='always'
    never='never'

@dataclass
class CommandlineParameterConfig:
    argument: str # default parameter for command line resutign ins "--argument"
    add_argument_args: dict # all arguments that will be passed to add_argument method
    short_form: Optional[str] = None
    group: str = 'misc'
    priority: int = 0
    in_call_string: InCallString = InCallString.auto

    def __gt__(self, other):
        """
        Sort required arguments first, then use priority, then name
        """
        return (not self.add_argument_args.get('required', False), -self.priority, self.argument)>(
            not other.add_argument_args.get('required', False), -other.priority, other.argument)

class ArgParsable:
    def __init_subclass__(cls):
        # create a nice documentation string that takes help into account
        cls.__doc__ = cls.__name__ + " Parameters:\n"
        for key, typ in cls.__annotations__.items():
            if get_origin(typ) is Union and type(None) in get_args(typ):
                optional = True
                typ = get_args(typ)[0]
            else:
                optional = False

            value = getattr(cls, key, None)
            try:
                cls.__doc__ += f"    {key} ({typ.__name__})"
            except AttributeError:
                cls.__doc__ += f"    {key}"
            if isinstance(value, Field):
                if value.default is not MISSING:
                    cls.__doc__ += f" = {value.default}"
                if 'help' in value.metadata:
                    cls.__doc__ += f" - {value.metadata['help']}"
            elif value is not None:
                    cls.__doc__ += f" = {value}"
            if optional:
                cls.__doc__ += " [Optional]"
            cls.__doc__ += "\n"
        return cls

    @classmethod
    def get_commandline_parameters(cls) -> List[CommandlineParameterConfig]:
        """
        Return a list of arguments used in building the command line parameters.

        Union types besides Optional are not supported.
        """
        output = []
        for field in fields(cls):
            args={}
            if field.default is not MISSING:
                args['default'] = field.default
                args['required'] = False
            elif field.default_factory is not MISSING:
                args['default'] = field.default_factory()
                args['required'] = False
            else:
                args['required'] = True
            if get_origin(field.type) is Union and type(None) in get_args(field.type):
                # optional argument
                typ = get_args(field.type)[0]
                del(args['default'])
            else:
                typ = field.type
            if get_origin(typ) is list:
                args['nargs'] = '+'
                args['action'] = 'extend'
                typ = get_args(typ)[0]
                if get_origin(typ) is tuple:
                    # tuple of items are put together during evaluation
                    typ = get_args(typ)[0]
            elif get_origin(typ) is tuple:
                args['nargs'] = len(get_args(typ))
                typ = get_args(typ)[0]
            if issubclass(typ, StrEnum):
                args['choices'] = [ci.value for ci in typ]
                if field.default is not MISSING:
                    args['default'] = field.default.value
                typ = str

            if typ is bool:
                args['action'] = 'store_false' if field.default else 'store_true'
            else:
                args['type'] = typ

            if 'help' in field.metadata:
                args['help'] = field.metadata['help']

            output.append(CommandlineParameterConfig(
                    field.name,
                    add_argument_args=args,
                    group=field.metadata.get('group', 'misc'),
                    short_form=field.metadata.get('short', None),
                    priority=field.metadata.get('priority', 0),
                    in_call_string=field.metadata.get('in_call_string', InCallString.auto),
                    ))
        return output

    @classmethod
    def get_default(cls, key):
        """
        Return the default argument for an attribute, None if it doesn't exist.
        """
        for field in fields(cls):
            if field.name != key:
                continue
            if field.default is not MISSING:
                return field.default
            elif field.default_factory is not MISSING:
                return field.default_factory()
        return None

    def is_default(self, key):
        value = getattr(self, key)
        return value == self.get_default(key)

    @classmethod
    def from_args(cls, args: argparse.Namespace):
        """
        Create the child class from the command line argument Namespace object.
        All attributes that are not needed for this class are ignored.
        """
        inpargs = {}
        for field in fields(cls):
            value = getattr(args, field.name)
            typ = field.type
            if get_origin(field.type) is Union and type(None) in get_args(field.type):
                # optional argument
                typ = get_args(field.type)[0]
            if get_origin(typ) is list:
                item_typ = get_args(typ)[0]
                if get_origin(item_typ) is tuple:
                    # tuple of items are put together during evaluation
                    tuple_length = len(get_args(item_typ))
                    value = [tuple(value[i*tuple_length+j] for j in range(tuple_length)) for i in range(len(value)//tuple_length)]
            if isinstance(typ, type) and issubclass(typ, StrEnum):
                # convert str to enum
                try:
                    value = typ(value)
                except ValueError:
                    choices = [ci.value for ci in typ]
                    raise ValueError(f"Parameter --{field.name} has to be one of {choices}")

            inpargs[field.name] = value
        return cls(**inpargs)

    def get_call_parameters(self, abbrv=True):
        """
        Return a list of command line arguments that reproduce this config, do not add default parameters.
        """
        output = []
        for arg in sorted(self.get_commandline_parameters()):
            if ((arg.in_call_string==InCallString.auto and self.is_default(arg.argument)) or
                    arg.in_call_string==InCallString.never):
                # skip default arguments or arguments defined to never appear in call string
                continue
            if arg.short_form and abbrv:
                item = '-' + arg.short_form + ' '
            else:
                item = '--' + arg.argument + ' '
            if arg.add_argument_args.get('type', None) in [str, float, int]:
                nargs = arg.add_argument_args.get('nargs', None)
                if nargs is None:
                    item += str(getattr(self, arg.argument))
                elif nargs=='+':
                    # remove the default parameters, only show added ones
                    ignore = len(arg.add_argument_args.get('default', []))
                    item += ' '.join([str(pi) for pi in getattr(self, arg.argument)[ignore:]])
                else:
                    item += ' '.join([str(pi) for pi in getattr(self, arg.argument)])
            # boolean flags only reach this point if they are non-default
            output.append((arg, item))
        return output

# definition of command line arguments

@dataclass
class ReaderConfig(ArgParsable):
    year: int = field(
            default=datetime.now().year,
            metadata={
                'short': 'Y', 
                'group': 'input data', 
                'help': 'year the measurement was performed',
                'in_call_string': InCallString.always,
                },
            )
    rawPath: List[str] = field(
            default_factory=lambda: ['.', path.join('.','raw'), path.join('..','raw'), path.join('..','..','raw')],
            metadata={
                'short': 'rp',
                'group': 'input data',
                'help': 'search paths for hdf files',
                },
            )
    startTime: Optional[float] = field(
            default = None,
            metadata={         
                'short': 'st',     
                'group': 'data manicure',
                'help': 'set time zero other than the start of the data aquisition',
                },
            )

class IncidentAngle(StrEnum):
    alphaF = 'alphaF'
    mu = 'mu'
    nu = 'nu'

class MonitorType(StrEnum):
    auto = 'a'
    proton_charge = 'p'
    time = 't'
    neutron_monitor = 'n'
    debug = 'x'

MONITOR_UNITS = {
    MonitorType.neutron_monitor: 'cnts',
    MonitorType.proton_charge: 'mC',
    MonitorType.time: 's',
    MonitorType.auto: 'various',
    MonitorType.debug: 'mC',
    }

@dataclass
class ExperimentConfig(ArgParsable):
    chopperPhase: float = field(
            default=0,
            metadata={
                'short': 'cp',
                'group': 'instrument settings',
                'help': 'phase between opening of chopper 1 and closing of chopper 2 window',
                },
            ) 
    chopperPhaseOffset: float = field(
            default=-5,
            metadata={
                'short': 'co',
                'group': 'instrument settings',  
                'help': 'phase between chopper 1 index pulse and closing edge',
                },                               
            )                          
    chopperSpeed: float = field(
            default=500,
            metadata={
                'short': 'cs',
                'group': 'instrument settings',
                'help': 'rotation speed of the chopper disks in rpm',
                },
            )
    yRange: Tuple[int, int] = field(
            default=(18, 48),
            metadata={
                'short': 'y',
                'group': 'region of interest',
                'help': 'horizontal pixel range on the detector to be used',
                },
            )
    lambdaRange: Tuple[float, float] = field(
            default_factory=lambda: [3, 12.5],
            metadata={
                'short': 'l',
                'group': 'region of interest',
                'help': 'wavelength range to be used (in angstrom)',
                },
            )
    lowCurrentThreshold: float = field(
            default=50,
            metadata={
                'short': 'pt',
                'group': 'instrument settings',
                'help': 'proton current below which the events are ignored (per chopper pulse)',
                },
            )

    incidentAngle: IncidentAngle = field(
            default=IncidentAngle.alphaF,
            metadata={
                'short': 'ai',
                'group': 'instrument settings',
                'help': 'calculate alphaI = [alphaF], [mu]+kappa+delta_kappa or ([nu]+kappa+delta_kappa)/2',
                },
            )
    alphaF = 'alphaF'
    sampleModel: Optional[str] = field(
            default=None,
            metadata={
                'short': 'sm',
                'group': 'sample',  
                'help': 'orso type string to describe the sample in one line',
                },                               
            )                          
    mu: Optional[float] = field(
            default=None,
            metadata={
                'short': 'mu',
                'group': 'sample',  
                'help': 'inclination of the sample surface w.r.t. the instrument horizon',
                },                               
            )                          
    nu: Optional[float] = field(
            default=None,
            metadata={
                'short': 'nu',
                'group': 'sample',  
                'help': 'inclination of the detector w.r.t. the instrument horizon',
                },                               
            )                          
    muOffset: Optional[float] = field(
            default=None,
            metadata={
                'short': 'm',
                'group': 'sample',  
                'help': 'correction offset for mu misalignment (mu_real = mu_file + mu_offset)',
                },                               
            )                          
    monitorType: MonitorType = field(
            default=MonitorType.proton_charge,
            metadata={
                'short': 'mt',
                'group': 'instrument settings',
                'help': 'one of [a]uto, [p]rotonCurrent, [t]ime or [n]eutronMonitor',
                },                               
            )                          

class NormalisationMethod(StrEnum):
    direct_beam = 'd'
    over_illuminated = 'o'
    under_illuminated = 'u'

@dataclass
class ReflectivityReductionConfig(ArgParsable):
    fileIdentifier: List[str] = field(
            metadata={
                'short':    'f',
                'priority': 100,
                'group':    'input data',
                'help':     'file number(s) or offset (if < 1)',
                },
            )

    qResolution: float = field(
            default=0.01,
            metadata={
                'short': 'r',
                'group': 'data manicure',
                'help': 'output resolution of q-scale Delta q / q',
                },
            )
    qzRange: Tuple[float, float] = field(
            default_factory=lambda: [0.005, 0.51],
            metadata={
                'short': 'q',
                'group': 'region of interest',
                'help': '?',
                },
            )
    thetaRange: Tuple[float, float] = field(
            default_factory=lambda: [-12., 12.],
            metadata={
                'short': 't',
                'group': 'region of interest',
                'help': 'absolute theta region of interest',
                },
            )
    thetaRangeR: Tuple[float, float] = field(
            default_factory=lambda: [-0.75, 0.75],
            metadata={
                'short': 'T',
                'group': 'region of interest',
                'help': 'theta region of interest w.r.t. beam center',
                },
            )
    thetaFilters: List[Tuple[float, float]] = field(
            default_factory=lambda: [],
            metadata={
                'short': 'TF',
                'group': 'region of interest',
                'help': 'add one or more theta ranges that will be filtered in reduction',
                },
            )
    normalisationMethod: NormalisationMethod = field(
            default=NormalisationMethod.over_illuminated,
            metadata={
                'short': 'nm', 
                'priority': 90,
                'group': 'input data',
                'help': 'normalisation method: [o]verillumination, [u]nderillumination, [d]irect_beam'})
    normalizationFilter: float = field(
            default=-1,
            metadata={
                'group': 'input data',
                'help': 'minimum normalization counts in lambda-theta bin to use, else filter'})
    normAngleFilter: float = field(
            default=-1,
            metadata={
                'group': 'input data',
                'help': 'minimum normalization counts total thetat bin to use, else filter'})
    normalizationSmoothing: float = field(
            default=0,
            metadata={
                'group': 'input data',
                'help': 'apply convolution on lambda axes to smooth the normalization data, useful for low statistics'})
    scale: List[float] = field(
            default_factory=lambda: [1.],
            metadata={
                'short': 's',
                'group': 'data manicure',
                'help': '(list of) scaling factors, if less elements than files use the last one',
                },
            ) 
    autoscale: Tuple[float, float] = field(
           default=None,
           metadata={
               'short': 'S',
               'group': 'data manicure',
               'help': '',
               },
           )
    subtract: Optional[str] = field(
            default=None, 
            metadata={
                'short': 'sub',
                'group': 'input data', 
                'help': 'File with R(q_z) curve to be subtracted (in .Rqz.ort format)'})
    normalisationFileIdentifier: Optional[List[str]] = field(
            default=None,
            metadata={
                'short': 'n', 
                'priority': 90,
                'group': 'input data', 
                'help': 'file number(s) of normalisation measurement'})
    timeSlize: Optional[List[float]] = field(
            default= None,
            metadata={
                'short': 'ts',
                'group': 'region of interest',
                'help': 'time slizing <interval> ,[<start> [,stop]]',
                },
            )

    logfilter: List[str] = field(
            default_factory=lambda: [],
            metadata={
                'short': 'lf',
                'group': 'region of interest',
                'help':  'filter using comparison to a log values, multpiple allowd (example "sample_temperature<150")',
                },
            )


class OutputFomatOption(StrEnum):
    Rqz_ort = "Rqz.ort"
    Rqz_orb = "Rqz.orb"
    Rlt_ort = "Rlt.ort"
    Rlt_orb = "Rlt.orb"
    ort = "ort"
    orb = "orb"
    Rqz = "Rqz"
    Rlt = "Rlt"


class PlotColormaps(StrEnum):
    gist_ncar = "gist_ncar"
    viridis = "viridis"
    inferno = "inferno"
    gist_rainbow = "gist_rainbow"
    nipy_spectral = "nipy_spectral"
    jochen_deluxe = "jochen_deluxe"

@dataclass
class ReflectivityOutputConfig(ArgParsable):
    outputFormats: List[OutputFomatOption] = field(
            default_factory=lambda: ['Rqz.ort'],
            metadata={
                'short': 'of',
                'group': 'output',
                'help': 'one of "Rqz[.ort]", "Rlt[.ort]" or both with "ort"',
                },
            )
    outputName: str = field(
            default='fromEOS',
            metadata={
                'short': 'o',
                'group': 'output',
                'help': '?',
                },
            )
    outputPath: str = field(
            default='.',
            metadata={
                'short': 'op',
                'group': 'output',
                'help': '?',
                },
            )
    plot: bool = field(
            default=False,
            metadata={
                'group': 'output',
                'help': 'show matplotlib graphs of results',
                },
            )

    plot_colormap: PlotColormaps = field(
            default=PlotColormaps.gist_ncar,
            metadata={
                'short': 'pcmap',
                'group': 'output',
                'help': 'matplotlib colormap used in lambda-theta graphs when plotting',
                },
            )

    append: bool = field(
            default=False,
            metadata={
                'group': 'output',
                'help': 'if file already exists, append result as additional ORSO dataset (only Rqz.ort)',
                },
            )

    def _output_format_list(self, outputFormat):
        format_list = []
        if OutputFomatOption.ort in outputFormat\
                or OutputFomatOption.Rqz_ort in outputFormat\
                or OutputFomatOption.Rqz in outputFormat:
            format_list.append(OutputFomatOption.Rqz_ort)
        if OutputFomatOption.ort in outputFormat\
                or OutputFomatOption.Rlt_ort in outputFormat\
                or OutputFomatOption.Rlt in outputFormat:
            format_list.append(OutputFomatOption.Rlt_ort)
        if OutputFomatOption.orb in outputFormat\
                or OutputFomatOption.Rqz_orb in outputFormat\
                or OutputFomatOption.Rqz in outputFormat:
            format_list.append(OutputFomatOption.Rqz_orb)
        if OutputFomatOption.orb in outputFormat\
                or OutputFomatOption.Rlt_orb in outputFormat\
                or OutputFomatOption.Rlt in outputFormat:
            format_list.append(OutputFomatOption.Rlt_orb)
        return sorted(format_list, reverse=True)

    def __post_init__(self):
        self.outputFormats = self._output_format_list(self.outputFormats)


# ===================================

@dataclass
class ReflectivityConfig:
    reader: ReaderConfig
    experiment: ExperimentConfig
    reduction: ReflectivityReductionConfig
    output: ReflectivityOutputConfig
    
    _call_string_overwrite=None
    
    #@property
    #def call_string(self)->str:
    #    if self._call_string_overwrite:
    #        return self._call_string_overwrite
    #    else:
    #        return self.calculate_call_string()
    
    def call_string(self):
        base = 'eos'

        call_parameters = self.reader.get_call_parameters()
        call_parameters += self.output.get_call_parameters()
        call_parameters += self.reduction.get_call_parameters()
        call_parameters += self.experiment.get_call_parameters()

        call_parameters.sort()

        cpout = f'{base} ' + ' '.join([cp[1] for cp in call_parameters])


        logging.debug(f'Argument list build in EOSConfig.call_string: {cpout}')
        return  cpout

class E2HPlotSelection(StrEnum):
    All = 'all'
    Raw = 'raw'
    YZ = 'Iyz'
    LT = 'Ilt'
    YT = 'Iyt'
    TZ = 'Itz'
    Q = 'Iq'
    L = 'Il'
    T = 'It'
    ToF = 'tof'


class E2HPlotArguments(StrEnum):
    Default = 'def'
    OutputFile = 'file'
    Logarithmic = 'log'
    Linear = 'lin'

@dataclass
class E2HReductionConfig(ArgParsable):
    fileIdentifier: str = field(
            default='0',
            metadata={
                'short':    'f',
                'priority': 100,
                'group':    'input data',
                'help':     'file number(s) or offset (if < 1), events2histogram only accepts one short code',
                },
            )

    show_plot: bool = field(
            default=False,
            metadata={
                'short': 'sp',
                'group': 'output',
                'help': 'show matplotlib graphs of results',
                },
            )

    plot: E2HPlotSelection = field(
            default=E2HPlotSelection.All,
            metadata={
                'short': 'p',
                'group': 'output',
                'help': 'select what to plot or write',
                },
            )

    kafka: bool = field(
            default=False,
            metadata={
                'group': 'output',
                'help': 'send result to kafka for Nicos',
                },
            )

    plotArgs: E2HPlotArguments = field(
            default=E2HPlotArguments.Default,
            metadata={
                'short': 'pa',
                'group': 'output',
                'help': 'select configuration for plot',
                },
            )

    update: bool = field(
            default=False,
            metadata={
                'short': 'u',
                'group': 'output',
                'help': 'keep running in the background and update plot when file is modified, implies --plot',
                },
            )

    fast: bool = field(
            default=False,
            metadata={
                'group': 'input data',
                'help': 'skip some reduction steps to speed up calculations',
                },
            )

    normalizationModel: bool = field(
            default=False,
            metadata={
                'short': 'nm',
                'group': 'input data',
                'help': 'use model for incoming spectrum and divergence to normalize for reflectivity',
                },
            )

    plot_colormap: PlotColormaps = field(
            default=PlotColormaps.jochen_deluxe,
            metadata={
                'short': 'pcmap',
                'group': 'output',
                'help': 'matplotlib colormap used in lambda-theta graphs when plotting',
                },
            )

    max_events: int = field(
            default = 10_000_000,
            metadata={
                'group': 'input data',
                'help':  'maximum number of events read at once',
                },
            )

    thetaRangeR: Tuple[float, float] = field(
            default_factory=lambda: [-0.75, 0.75],
            metadata={
                'short': 'T',
                'group': 'region of interest',
                'help': 'theta region of interest w.r.t. beam center',
                },
            )

    thetaFilters: List[Tuple[float, float]] = field(
            default_factory=lambda: [],
            metadata={
                'short': 'TF',
                'group': 'region of interest',
                'help': 'add one or more theta ranges that will be filtered in reduction',
                },
            )

    fontsize: float = field(
            default=8.,
            metadata={
                'short': 'pf',
                'group': 'output',
                'help': 'font size for graphs',
                },
            )

@dataclass
class E2HConfig:
    reader: ReaderConfig
    experiment: ExperimentConfig
    reduction: E2HReductionConfig



"""
Equivalent function as in helpers_numba.py but using just numpy functionality.
"""

import numpy as np

def merge_frames(tof_e, tofCut, tau, total_offset):
    # tof shifted to 1 frame
    return np.remainder(tof_e-(tofCut-tau), tau)+total_offset

def extract_walltime(tof_e, dataPacket_p, dataPacketTime_p):
    output = np.empty(np.shape(tof_e)[0], dtype=np.int64)
    for i in range(len(dataPacket_p)-1):
        output[dataPacket_p[i]:dataPacket_p[i+1]] = dataPacketTime_p[i]
    output[dataPacket_p[-1]:] = dataPacketTime_p[-1]
    return output

def filter_project_x(pixelLookUp, pixelID_e, ymin, ymax):
    (detY_e, detZ_e, detXdist_e, delta_e) = pixelLookUp[np.int_(pixelID_e)-1, :].T
    # define mask and filter y range
    mask_e = (ymin<=detY_e) & (detY_e<=ymax)
    return (detZ_e, detXdist_e, delta_e, mask_e)

def calculate_derived_properties_focussing(tof_e, detXdist_e, delta_e, mask_e,
                                           lmin, lmax, nu, mu, chopperDetectorDistance, hdm):
    raise NotImplementedError("Only exists in numba implementation so far.")



"""
Reading of Amor NeXus data files to extract metadata and event stream.
"""
from typing import BinaryIO, List, Union

import sys
import h5py
import numpy as np
import logging
import subprocess

from datetime import datetime

from orsopy import fileio
from orsopy.fileio.model_language import SampleModel

from . import const
from .header import Header
from .event_data_types import AmorGeometry, AmorTiming, AmorEventStream, LOG_TYPE, PACKET_TYPE, EVENT_TYPE, PULSE_TYPE, \
    PC_TYPE

try:
    import zoneinfo
except ImportError:
    # for python versions < 3.9 try to use the backports version
    from backports import zoneinfo


# Time zone used to interpret time strings
AMOR_LOCAL_TIMEZONE = zoneinfo.ZoneInfo(key='Europe/Zurich')
UTC = zoneinfo.ZoneInfo(key='UTC')

class AmorHeader:
    """
    Collects header information from Amor NeXus fiel without reading event data.
    """
    # mapping of names to (hdf_path, dtype, nicos_key[, suffix])
    hdf_paths = dict(
            title=('entry1/title', str),
            proposal_id=('entry1/proposal_id', str),
            user_name=('entry1/user/name', str),
            user_email=('entry1/user/email', str),
            sample_name=('entry1/sample/name', str),
            source_name=('entry1/Amor/source/name', str),
            sample_model=('entry1/sample/model', str),
            start_time=('entry1/start_time', str),
            start_time_fallback=('entry1/Amor/instrument_control_parameters/start_time', str),

            chopper_separation=('entry1/Amor/chopper/pair_separation', float),
            detector_distance=('entry1/Amor/detector/transformation/distance', float),
            chopper_distance=('entry1/Amor/chopper/distance', float),
            sample_temperature=('entry1/sample/temperature', float),
            sample_magnetic_field=('entry1/sample/magnetic_field', float),

            mu=('entry1/Amor/instrument_control_parameters/mu', float, 'mu'),
            nu=('entry1/Amor/instrument_control_parameters/nu', float, 'nu'),
            kap=('entry1/Amor/instrument_control_parameters/kappa', float, 'kappa'),
            kad=('entry1/Amor/instrument_control_parameters/kappa_offset', float, 'kappa_offset'),
            div=('entry1/Amor/instrument_control_parameters/div', float, 'div'),
            ch1_trigger_phase=('entry1/Amor/chopper/ch1_trigger_phase', float, 'ch1_trigger_phase'),
            ch2_trigger_phase=('entry1/Amor/chopper/ch2_trigger_phase', float, 'ch2_trigger_phase'),
            chopper_speed=('entry1/Amor/chopper/rotation_speed', float, 'chopper_phase'),
            chopper_phase=('entry1/Amor/chopper/phase', float, 'chopper_phase'),
            polarization_config_label=('entry1/Amor/polarization/configuration', int, 'polarization_config_label', '/*'),
            )

    def __init__(self, fileName:Union[str, h5py.File, BinaryIO]):
        if type(fileName) is str:
            logging.warning(f'    {fileName.split("/")[-1]}')
            self.hdf = h5py.File(fileName, 'r', swmr=True)
        elif type(fileName) is h5py.File:
            self.hdf = fileName
        else:
            self.hdf = h5py.File(fileName, 'r')

        self._log_keys = []

        self.read_header_info()
        self.read_instrument_configuration()

        if type(fileName) is str:
            # close the input file to free memory, only if the file was opened in this object
            self.hdf.close()
        del(self.hdf)

    def _replace_if_missing(self, key, nicos_key, dtype=float, suffix=''):
        from .nicos_interface import lookup_nicos_value
        year = self.fileDate.strftime('%Y')
        return lookup_nicos_value(key, nicos_key, dtype, suffix, year)

    def rv(self, key):
        """
        Generic read value methos based on key in hdf_paths dictionary.
        """
        hdf_path, dtype, *nicos = self.hdf_paths[key]
        try:
            hdfgrp = self.hdf[hdf_path]
            if hdfgrp.attrs.get('NX_class', None) == 'NXlog':
                # use the last value that was recoreded before the count started
                time_column = hdfgrp['time'][:]
                try:
                    start_index = np.where(time_column<=self._start_time_ns)[0][0]
                except IndexError:
                    start_index = 0
                if hdfgrp['value'].ndim==1:
                    output = dtype(hdfgrp['value'][start_index])
                else:
                    output = dtype(hdfgrp['value'][start_index, 0])
                # make sure key is only appended if no exception was raised
                self._log_keys.append(key)
                return output
            elif dtype is str:
                return self.read_string(hdf_path)
            else:
                if len(hdfgrp.shape)==1:
                    return dtype(hdfgrp[0])
                else:
                    return dtype(hdfgrp[()])
        except (KeyError, IndexError):
            if nicos:
                nicos_key = nicos[0]
                suffix = nicos[1] if len(nicos)>1 else ''
                return self._replace_if_missing(key, nicos_key, dtype, suffix)
            else:
                raise


    def read_string(self, path):
        if not np.shape(self.hdf[path]):
            return self.hdf[path][()].decode('utf-8')
        else:
            # format until 2025
            return self.hdf[path][0].decode('utf-8')

    def read_header_info(self):
        self._start_time_ns = np.uint64(0)
        try:
            start_time = self.rv('start_time')
        except KeyError:
            start_time = self.rv('start_time_fallback')

        # extract start time as unix time, adding UTC offset of 1h to time string
        if start_time.endswith('Z') and sys.version_info.minor<11:
            # older python versions did not support Z format
            start_time = start_time[:-1]
            TZ = UTC
        else:
            TZ = AMOR_LOCAL_TIMEZONE
        start_date = datetime.fromisoformat(start_time)
        self.fileDate = start_date.replace(tzinfo=TZ)
        self._start_time_ns = np.uint64(self.fileDate.timestamp()*1e9)

        # read general information and first data set
        title = self.rv('title')
        proposal_id = self.rv('proposal_id')
        user_name = self.rv('user_name')
        user_affiliation = 'unknown'
        user_email = self.rv('user_email')
        user_orcid = None
        sampleName = self.rv('sample_name')
        instrumentName = 'Amor'
        source = self.rv('source_name')
        sourceProbe = 'neutron'
        model = self.rv('sample_model')
        if 'stack' in model:
            import yaml
            model = yaml.safe_load(model)
        else:
            model = dict(stack=model)

        self.owner = fileio.Person(
                name=user_name,
                affiliation=user_affiliation,
                contact=user_email,
                )
        if user_orcid:
            self.owner.orcid = user_orcid

        self.experiment = fileio.Experiment(
                title=title,
                instrument=instrumentName,
                start_date=start_date,
                probe=sourceProbe,
                facility=source,
                proposalID=proposal_id
                )
        if model['stack'] == '':
            om = None
        else:
            om = SampleModel.from_dict(model)
        self.sample = fileio.Sample(
                name=sampleName,
                model=om,
                sample_parameters={},
                )
        # while event times are not evaluated, use average_value reported in file for SEE
        if self.hdf['entry1/sample'].get('temperature', None) is not None:
            try:
                sample_temperature = self.rv('sample_temperature')
            except IndexError: pass
            else:
                self.sample.sample_parameters['temperature'] = fileio.Value(sample_temperature, unit='K')
        if self.hdf['entry1/sample'].get('magnetic_field', None) is not None:
            try:
                sample_magnetic_field = self.rv('sample_magnetic_field')
            except IndexError: pass
            else:
                self.sample.sample_parameters['magnetic_field'] = fileio.Value(sample_magnetic_field, unit='T')

    def read_instrument_configuration(self):
        chopperSeparation = self.rv('chopper_separation')
        detectorDistance = self.rv('detector_distance')
        chopperDistance = self.rv('chopper_distance')
        chopperDetectorDistance = detectorDistance - chopperDistance

        mu = self.rv('mu')
        nu = self.rv('nu')
        kap = self.rv('kap')
        kad = self.rv('kad')
        div = self.rv('div')
        ch1TriggerPhase = self.rv('ch1_trigger_phase')
        ch2TriggerPhase = self.rv('ch2_trigger_phase')
        try:
            chopperTriggerTime = (float(self.hdf['entry1/Amor/chopper/ch2_trigger/event_time_zero'][7]) \
                                  -float(self.hdf['entry1/Amor/chopper/ch2_trigger/event_time_zero'][0])) \
                                 /7
            chopperTriggerTimeDiff = float(self.hdf['entry1/Amor/chopper/ch2_trigger/event_time_offset'][2])
        except (KeyError, IndexError):
            logging.debug('      chopper speed and phase taken from .hdf file')
            chopperSpeed = self.rv('chopper_speed')
            chopperPhase = self.rv('chopper_phase')
            tau = 30/chopperSpeed
        else:
            tau = int(1e-6*chopperTriggerTime/2+0.5)*(1e-3)
            chopperTriggerPhase = 180e-9*chopperTriggerTimeDiff/tau
            chopperSpeed = 30/tau
            chopperPhase = chopperTriggerPhase+ch1TriggerPhase-ch2TriggerPhase

        self.geometry = AmorGeometry(mu, nu, kap, kad, div,
                                     chopperSeparation, detectorDistance, chopperDetectorDistance)
        self.timing = AmorTiming(ch1TriggerPhase, ch2TriggerPhase, chopperSpeed, chopperPhase, tau)

        polarizationConfigLabel = self.rv('polarization_config_label')
        polarizationConfig = fileio.Polarization(const.polarizationConfigs[polarizationConfigLabel])
        logging.debug(f'      polarization configuration: {polarizationConfig} (index {polarizationConfigLabel})')


        self.instrument_settings = fileio.InstrumentSettings(
            incident_angle = fileio.ValueRange(round(mu+kap+kad-0.5*div, 3),
                                               round(mu+kap+kad+0.5*div, 3),
                                               'deg'),
            wavelength = fileio.ValueRange(const.lamdaCut, const.lamdaMax, 'angstrom'),
            polarization = fileio.Polarization(polarizationConfig)
            )
        self.instrument_settings.qz = fileio.ValueRange(round(4*np.pi*np.sin(np.deg2rad(mu+kap+kad-0.5*div))/const.lamdaMax, 3),
                                                        round(4*np.pi*np.sin(np.deg2rad(mu+kap+kad+0.5*div))/const.lamdaCut, 3),
                                                        '1/angstrom')
        self.instrument_settings.mu = fileio.Value(
                round(mu, 3),
                'deg',
                comment='sample angle to horizon')
        self.instrument_settings.nu = fileio.Value(
                round(nu, 3),
                'deg',
                comment='detector angle to horizon')
        self.instrument_settings.div = fileio.Value(
                round(div, 3),
                'deg',
                comment='incoming beam divergence')
        self.instrument_settings.kap = fileio.Value(
                round(kap, 3),
                'deg',
                comment='incoming beam inclination')
        if abs(kad)>0.02:
            self.instrument_settings.kad = fileio.Value(
                    round(kad, 3),
                    'deg',
                    comment='incoming beam angular offset')


    def update_header(self, header:Header):
        """
        Add dataset information into an existing header.
        """
        logging.info(f'    meta data from: {self.file_list[0]}')
        header.owner = self.owner
        header.experiment = self.experiment
        header.sample = self.sample
        header.measurement_instrument_settings = self.instrument_settings


class AmorEventData(AmorHeader):
    """
    Read one amor NeXus datafile and extract relevant header information.

    Implements EventDatasetProtocol
    """
    file_list: List[str]
    first_index: int
    last_index: int = -1
    EOF: bool = False
    max_events: int
    owner: fileio.Person
    experiment: fileio.Experiment
    sample: fileio.Sample
    instrument_settings: fileio.InstrumentSettings
    geometry: AmorGeometry
    timing: AmorTiming
    data: AmorEventStream

    eventStartTime: np.int64

    def __init__(self, fileName:Union[str, h5py.File, BinaryIO], first_index:int=0, max_events:int=100_000_000):
        if type(fileName) is str:
            logging.warning(f'    {fileName.split("/")[-1]}')
            self.file_list = [fileName]
            hdf = h5py.File(fileName, 'r', swmr=True)
        elif type(fileName) is h5py.File:
            self.file_list = [fileName.filename]
            hdf = fileName
        else:
            self.file_list = [repr(fileName)]
            hdf = h5py.File(fileName, 'r')
        self.first_index = first_index
        self.max_events = max_events

        super().__init__(hdf)
        self.hdf = hdf
        try:
            self.read_event_stream()
        except EOFError:
            self.hdf.close()
            del(self.hdf)
            raise
        self.read_log_streams()

        if type(fileName) is str:
            # close the input file to free memory, only if the file was opened in this object
            self.hdf.close()
        del(self.hdf)


    def read_event_stream(self):
        """
        Read the actual event data from file. If file is too large, find event index from packets
        that allow splitting of file smaller than self.max_events.
        """
        packets = np.recarray(self.hdf['/entry1/Amor/detector/data/event_index'].shape, dtype=PACKET_TYPE)
        packets.start_index = self.hdf['/entry1/Amor/detector/data/event_index'][:]
        packets.time = self.hdf['/entry1/Amor/detector/data/event_time_zero'][:]
        try:
            # packet index that matches first event index
            start_packet = int(np.where(packets.start_index==self.first_index)[0][0])
        except IndexError:
            raise EOFError(f'No event packet found starting at event #{self.first_index}, '
                           f'number of events is {self.hdf["/entry1/Amor/detector/data/event_time_offset"].shape[0]}')
        packets = packets[start_packet:]
        if packets.shape[0]==0:
            raise EOFError(f'No more packets left after start_packet filter')

        nevts = self.hdf['/entry1/Amor/detector/data/event_time_offset'].shape[0]
        if (nevts-self.first_index)>self.max_events:
            end_packet = np.where(packets.start_index<=(self.first_index+self.max_events))[0][-1]
            end_packet = max(1, end_packet)
            if len(packets)==1:
                self.last_index = nevts-1
            else:
                self.last_index = packets.start_index[end_packet]-1
            packets = packets[:end_packet]
        else:
            self.last_index = nevts-1
            self.EOF = True

        if packets.shape[0]==0:
            raise EOFError(f'No more packets left after end_packet filter, first_index={self.first_index}, '
                           f'max_events={self.max_events}, nevts={nevts}')

        nevts = self.last_index+1-self.first_index

        # adapte packet to event index relation
        packets.start_index -= self.first_index

        events = np.recarray(nevts, dtype=EVENT_TYPE)
        events.tof = np.array(self.hdf['/entry1/Amor/detector/data/event_time_offset'][self.first_index:self.last_index+1])/1.e9
        events.pixelID = self.hdf['/entry1/Amor/detector/data/event_id'][self.first_index:self.last_index+1]
        events.mask = 0

        pulses = self.read_chopper_trigger_stream(packets)
        current = self.read_proton_current_stream(packets)
        self.data = AmorEventStream(events, packets, pulses, current)

        if self.first_index>0 and not self.EOF:
            # label the file name if not all events were used
            self.file_list[0] += f'[{self.first_index}:{self.last_index}]'

    def read_log_streams(self):
        """
        Read the individual NXlog datasets that can later be used for filtering etc.
        """
        for key in self._log_keys:
            hdf_path, dtype, *_ = self.hdf_paths[key]
            hdfgroup = self.hdf[hdf_path]
            shape = hdfgroup['time'].shape
            data = np.recarray(shape, dtype=np.dtype([('value', self.hdf_paths[key][1]), ('time', np.int64)]))
            data.time = hdfgroup['time'][:]
            if len(hdfgroup['value'].shape)==1:
                data.value = hdfgroup['value'][:]
            else:
                data.value = hdfgroup['value'][:, 0]
            self.data.device_logs[key] = data

    def update_info_from_logs(self):
        RELEVANT_ITEMS = ['sample_temperature', 'sample_magnetic_field', 'polarization_config_label']
        for key, log in self.data.device_logs.items():
            if key not in RELEVANT_ITEMS:
                continue
            if log.value.dtype in [np.int8, np.int16, np.int32, np.int64]:
                # for integer items (flags) report the most common one
                value = np.bincount(log.value).argmax()
                if logging.getLogger().getEffectiveLevel() <= logging.DEBUG \
                        and np.unique(log.value).shape[0]>1:
                    logging.debug(f'        filtered values for {key} not unique, '
                                  f'has {np.unique(log.value).shape[0]} values')
            else:
                value = log.value.mean()
            if key == 'polarization_config_label':
                self.instrument_settings.polarization = fileio.Polarization(const.polarizationConfigs[value])
            elif key == 'sample_temperature':
                self.sample.sample_parameters['temperature'].magnitue = value
            elif key == 'sample_magnetic_field':
                self.sample.sample_parameters['magnetic_field'].magnitue = value



    def read_chopper_trigger_stream(self, packets):
        chopper1TriggerTime = np.array(self.hdf['entry1/Amor/chopper/ch2_trigger/event_time_zero'][:-2], dtype=np.int64)
        #self.chopper2TriggerTime = self.chopper1TriggerTime + np.array(self.hdf['entry1/Amor/chopper/ch2_trigger/event_time'][:-2], dtype=np.int64)
        #                           + np.array(self.hdf['entry1/Amor/chopper/ch2_trigger/event_time_offset'][:], dtype=np.int64)
        if np.shape(chopper1TriggerTime)[0] > 2:
            startTime = chopper1TriggerTime[0]
            pulseTimeS = chopper1TriggerTime
        else:
            logging.critical('     No chopper trigger data available, using event steram instead, pulse filtering will fail!')
            startTime = np.array(self.hdf['/entry1/Amor/detector/data/event_time_zero'][0], dtype=np.int64)
            stopTime = np.array(self.hdf['/entry1/Amor/detector/data/event_time_zero'][-2], dtype=np.int64)
            pulseTimeS = np.arange(startTime, stopTime, self.timing.tau*1e9, dtype=np.int64)
        pulses = np.recarray(pulseTimeS.shape, dtype=PULSE_TYPE)
        pulses.time = pulseTimeS
        pulses.monitor = 1. # default is monitor pulses as it requires no calculation
        # apply filter in case the events were filtered
        if (self.first_index>0 or not self.EOF):
            pulses = pulses[(pulses.time>=packets.time[0])&(pulses.time<=packets.time[-1])]
        self.eventStartTime = startTime
        return pulses

    def read_proton_current_stream(self, packets):
        proton_current = np.recarray(self.hdf['entry1/Amor/detector/proton_current/time'].shape, dtype=PC_TYPE)
        proton_current.time = self.hdf['entry1/Amor/detector/proton_current/time'][:]
        if self.hdf['entry1/Amor/detector/proton_current/value'].ndim==1:
            proton_current.current = self.hdf['entry1/Amor/detector/proton_current/value'][:]
        else:
            proton_current.current = self.hdf['entry1/Amor/detector/proton_current/value'][:,0]

        if self.first_index>0 or not self.EOF:
            proton_current = proton_current[(proton_current.time>=packets.time[0])&
                                            (proton_current.time<=packets.time[-1])]
        return proton_current

    def info(self):
        output = ""
        for key in ['owner', 'experiment', 'sample', 'instrument_settings']:
            value = repr(getattr(self, key)).replace("\n","\n      ")
            output += f'\n{key}={value},'
        output += '\n'
        return output

    def append(self, other):
        """
        Append event streams from another file to this one. Adjusts the event indices in the
        packets to stay valid.
        """
        new_events = np.concatenate([self.data.events, other.data.events]).view(np.recarray)
        new_pulses = np.concatenate([self.data.pulses, other.data.pulses]).view(np.recarray)
        new_proton_current = np.concatenate([self.data.proton_current, other.data.proton_current]).view(np.recarray)
        new_packets = np.concatenate([self.data.packets, other.data.packets]).view(np.recarray)
        new_packets.start_index[self.data.packets.shape[0]:] += self.data.events.shape[0]
        self.data = AmorEventStream(new_events, new_packets, new_pulses, new_proton_current)
        # Indicate that this is amodified dataset, basically counts number of appends as negative indices
        self.last_index = min(self.last_index-1, -1)
        self.file_list += other.file_list

    def __repr__(self):
        output = (f"AmorEventData({self.file_list!r}) # {self.data.events.shape[0]} events, "
                  f"{self.data.pulses.shape[0]} pulses")

        return output

    def get_timeslice(self, start, end)->'AmorEventData':
        # return a new dataset with just events that occured in given time slice
        if not 'wallTime' in self.data.events.dtype.names:
            raise ValueError("This dataset is missing a wallTime that is required for time slicing")
        # convert from seconds to epoch integer values
        start , end = start*1e9, end*1e9
        event_filter = self.data.events.wallTime>=start
        event_filter &= self.data.events.wallTime<end
        pulse_filter = self.data.pulses.time>=start
        pulse_filter &= self.data.pulses.time<end
        output = super().__new__(AmorEventData)
        for key, value in self.__dict__.items():
            if key == 'data':
                continue
            else:
                setattr(output, key, value)
        # TODO: this is not strictly correct, as the packet/event relationship is lost
        output.data = AmorEventStream(self.data.events[event_filter], self.data.packets,
                                      self.data.pulses[pulse_filter], self.data.proton_current)
        return output



"""
Define an event dataformat that performs reduction actions like wavelength calculation on per-event basis.
With large number of events these actions can be time consuming so they use numba based functions.
"""
import numpy as np
import logging

from typing import Tuple

from . import const
from .event_data_types import EventDataAction, EventDatasetProtocol, append_fields, EVENT_BITMASKS
from .helpers import filter_project_x, merge_frames, extract_walltime, add_log_to_pulses
from .instrument import Detector
from .options import IncidentAngle
from .header import Header

class ExtractWalltime(EventDataAction):
    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        wallTime = extract_walltime(dataset.data.events.tof,
                                    dataset.data.packets.start_index,
                                    dataset.data.packets.time)
        logging.debug(f'        expending event stream by wallTime')
        new_events = append_fields(dataset.data.events, [('wallTime', wallTime.dtype)])
        new_events.wallTime = wallTime
        dataset.data.events = new_events

class MergeFrames(EventDataAction):
    def __init__(self, lamdaCut=None):
        self.lamdaCut=lamdaCut

    def perform_action(self, dataset: EventDatasetProtocol)->None:
        if self.lamdaCut is None:
            lamdaCut = const.lamdaCut
        else:
            lamdaCut = self.lamdaCut
        tofCut = lamdaCut*dataset.geometry.chopperDetectorDistance/const.hdm*1e-13
        total_offset = (tofCut +
                        dataset.timing.tau * (dataset.timing.ch1TriggerPhase + dataset.timing.chopperPhase/2)/180)
        dataset.data.events.tof = merge_frames(dataset.data.events.tof, tofCut, dataset.timing.tau, total_offset)


class AnalyzePixelIDs(EventDataAction):
    def __init__(self, yRange: Tuple[int, int]):
        self.yRange = yRange

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        d = dataset.data
        (detZ, detXdist, delta, mask) = filter_project_x(
                Detector.pixelLookUp, d.events.pixelID, self.yRange[0], self.yRange[1]
                )
        ana_events = append_fields(d.events, [
            ('detZ', detZ.dtype), ('detXdist', detXdist.dtype), ('delta', delta.dtype)])
        # add analysis per event
        ana_events.detZ = detZ
        ana_events.detXdist = detXdist
        ana_events.delta = delta
        ana_events.mask += np.logical_not(mask)*EVENT_BITMASKS['yRange']
        d.events = ana_events

class CalculateWavelength(EventDataAction):
    def __init__(self, lambdaRange: Tuple[float, float]):
        self.lambdaRange = lambdaRange

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        d = dataset.data
        if not 'detXdist' in dataset.data.events.dtype.names:
            raise ValueError("CalculateWavelength requires dataset with analyzed pixels, perform AnalyzePixelIDs first")

        #lamdaMax = const.lamdaCut+1.e13*dataset.timing.tau*const.hdm/(dataset.geometry.chopperDetectorDistance+124.)

        # lambda
        # TODO: one of the most time consuming actions, could be implemented in numba, instead?
        lamda = (1.e13*const.hdm)*d.events.tof/(dataset.geometry.chopperDetectorDistance+d.events.detXdist)

        final_events = append_fields(d.events, [('lamda', np.float64)])
        # add analysis per event
        final_events.lamda = lamda
        final_events.mask += EVENT_BITMASKS["LamdaRange"]*(
                (self.lambdaRange[0]>lamda) | (lamda>self.lambdaRange[1]))
        d.events = final_events

class CalculateQ(EventDataAction):
    def __init__(self, incidentAngle: IncidentAngle):
        self.incidentAngle = incidentAngle

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        d = dataset.data
        if not 'lamda' in dataset.data.events.dtype.names:
            raise ValueError("CalculateQ requires dataset with analyzed wavelength, perform CalculateWavelength first")

        lamda = d.events.lamda

        final_events = append_fields(d.events, [('qz', np.float64)])

        # alpha_f
        # q_z
        if self.incidentAngle == IncidentAngle.alphaF:
            alphaF_e  = dataset.geometry.nu - dataset.geometry.mu + d.events.delta
            final_events.qz = 4*np.pi*(np.sin(np.deg2rad(alphaF_e))/lamda)
        elif self.incidentAngle == IncidentAngle.nu:
            alphaF_e  = (dataset.geometry.nu + d.events.delta + dataset.geometry.kap + dataset.geometry.kad) / 2.
            final_events.qz = 4*np.pi*(np.sin(np.deg2rad(alphaF_e))/lamda)
        else:
            alphaF_e  = dataset.geometry.nu - dataset.geometry.mu + d.events.delta
            alphaI    = dataset.geometry.kap + dataset.geometry.kad + dataset.geometry.mu
            final_events.qz = 2*np.pi * ((np.sin(np.deg2rad(alphaF_e)) + np.sin(np.deg2rad(alphaI)))/lamda)
            final_events = append_fields(final_events, [('qx', np.float64)])
            final_events.qx = 2*np.pi * ((np.cos(np.deg2rad(alphaF_e)) - np.cos(np.deg2rad(alphaI)))/lamda)

        dataset.data.events = final_events

    def update_header(self, header: Header):
        if self.incidentAngle == IncidentAngle.alphaF:
            header.measurement_scheme = 'angle- and energy-dispersive'
        else:
            header.measurement_scheme = 'energy-dispersive'

class FilterQzRange(EventDataAction):
    def __init__(self, qzRange: Tuple[float, float]):
        self.qzRange = qzRange

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        d = dataset.data
        if not 'qz' in dataset.data.events.dtype.names:
            raise ValueError("FilterQzRange requires dataset with qz values per events, perform WavelengthAndQ first")

        d.events.mask += EVENT_BITMASKS["qRange"]*((self.qzRange[0]>d.events.qz) | (d.events.qz>self.qzRange[1]))

class FilterByLog(EventDataAction):

    def __init__(self, filter_string, remove_switchpulse=False):
        if filter_string.startswith('!'):
            filter_string = filter_string[1:]
            remove_switchpulse = True
        self.filter_string = filter_string
        self.remove_switchpulse = remove_switchpulse

    def perform_action(self, dataset: EventDatasetProtocol) -> None:
        filter_variable = None
        # go through existing keys in reverse order of their length to insure a name containing another is used
        existing_keys = list(dataset.data.device_logs.keys())
        existing_keys.sort(key=lambda x: -len(x))
        for key in existing_keys:
            if key in self.filter_string:
                filter_variable = key
                break
        if filter_variable is None:
            logging.warning(f'          could not find suitable parameter to filter on in {self.filter_string}, '
                            f'available parameters are: {list(sorted(existing_keys))}')
            return
        logging.debug(f'          using parameter {filter_variable} to apply filter {self.filter_string}')
        if not filter_variable in EVENT_BITMASKS:
            EVENT_BITMASKS[filter_variable] = max(EVENT_BITMASKS.values())*2
        if not filter_variable in dataset.data.pulses.dtype.names:
            # interpolate the parameter values for all existing pulses
            add_log_to_pulses(filter_variable, dataset)
        fltr_pulses = eval(self.filter_string, {filter_variable: dataset.data.pulses[filter_variable]})
        if self.remove_switchpulse:
            switched = fltr_pulses[:-1] & ~fltr_pulses[1:]
            fltr_pulses[:-1] &= ~switched
        goodTimeS = dataset.data.pulses.time[fltr_pulses]
        filter_e = np.logical_not(np.isin(dataset.data.events.wallTime, goodTimeS))
        dataset.data.events.mask += EVENT_BITMASKS[filter_variable]*filter_e



"""
Class to handle Orso header information that changes gradually during the reduction process.
"""

import platform
import sys
from datetime import datetime
from typing import List, Literal

from orsopy import fileio

from . import __version__


class Header:
    """orso compatible output file header content"""
    owner: fileio.Person
    experiment: fileio.Experiment
    sample: fileio.Sample
    measurement_instrument_settings: fileio.InstrumentSettings
    measurement_scheme: Literal["angle- and energy-dispersive", "angle-dispersive", "energy-dispersive"]
    measurement_data_files: List[fileio.File]
    measurement_additional_files: List[fileio.File]


    def __init__(self):
        self.measurement_data_files          = []
        self.measurement_additional_files    = []

        self.reduction = fileio.Reduction(
            software    = fileio.Software('eos', version=__version__),
            call        = 'placeholder',
            computer    = platform.node(),
            timestamp   = datetime.now(),
            creator     = None,
            corrections = ['histogramming in lambda and alpha_f',
                           'gravity'],
            )
    #-------------------------------------------------------------------------------------------------
    def data_source(self):
        return fileio.DataSource(
            self.owner,
            self.experiment,
            self.sample,
            fileio.Measurement(
                instrument_settings = self.measurement_instrument_settings,
                scheme              = self.measurement_scheme,
                data_files          = self.measurement_data_files,
                additional_files    = self.measurement_additional_files,
                ),
        )
    #-------------------------------------------------------------------------------------------------
    def columns(self):
        cols = [
            fileio.Column('Qz', '1/angstrom', 'normal momentum transfer'),
            fileio.Column('R', '', 'specular reflectivity'),
            fileio.ErrorColumn(error_of='R', error_type='uncertainty', distribution='gaussian', value_is='sigma'),
            fileio.ErrorColumn(error_of='Qz', error_type='resolution', distribution='gaussian', value_is='sigma'),
            ]
        return cols
    #-------------------------------------------------------------------------------------------------
    def orso_header(self, columns=None, extra_columns=[]):
        """
        Generate ORSO header from a copy of this class' data.
        """
        ds = fileio.DataSource.from_dict(self.data_source().to_dict())
        red = fileio.Reduction.from_dict(self.reduction.to_dict())
        if columns is None:
            columns = self.columns()
        return fileio.Orso(ds, red, columns+extra_columns)



"""
eosls executable script to list available datafiles in current folder with some metadata information.

Author: Jochen Stahn (algorithms, python draft),
        Artur Glavic (structuring and optimisation of code)
"""
import os
import logging

from eos.command_line import commandLineArgs

def main():
    logging.getLogger().setLevel(logging.CRITICAL)
    clas = commandLineArgs([], 'eosls', extra_args=[
        dict(dest='path', nargs='*', default=['.'], help='paths to list file in')])

    from glob import glob
    import tabulate
    from eos.file_reader import AmorHeader

    files = []
    for path in clas.path:
        files+=glob(os.path.join(path, 'amor*.hdf'))
    files.sort()

    data = {
        'File name': [],
        'Start Time': [],
        'mu': [],
        'nu': [],
        'div': [],
        'Sample': [],
        'T [K]': [],
        'H [T]': [],
        }
    for fi in files:
        data['File name'].append(os.path.basename(fi))
        ah = AmorHeader(fi)
        data['Sample'].append(ah.sample.name)
        data['Start Time'].append(ah.fileDate.strftime('%y %m-%d %H:%M:%S'))
        data['mu'].append('%.3f' % ah.geometry.mu)
        data['nu'].append('%.3f' % ah.geometry.nu)
        data['div'].append('%.3f' % ah.geometry.div)

        T = ah.sample.sample_parameters.get('temperature', None)
        data['T [K]'].append(T.magnitude if T is not None else '-')

        H = ah.sample.sample_parameters.get('magnetic_field', None)
        data['H [T]'].append(H.magnitude if H is not None else '-')

    print(tabulate.tabulate(data, headers="keys"))

if __name__ == '__main__':
    main()



"""
Defines how file paths are resolved from short_notation, year and number to filename.
"""
import logging
import os
from typing import List


class PathResolver:
    def __init__(self, year, rawPath):
        self.year = year
        self.rawPath = rawPath

    def resolve(self, short_notation):
        return list(map(self.get_path, self.expand_file_list(short_notation)))

    @staticmethod
    def expand_file_list(short_notation)->List[int]:
        """Evaluate string entry for file number lists"""
        file_list = []
        for i in short_notation.split(','):
            if '-' in i and not i.startswith('-'):
                if ':' in i:
                    step = i.split(':', 1)[1]
                    file_list += range(int(i.split('-', 1)[0]),
                                       int((i.rsplit('-', 1)[1]).split(':', 1)[0])+1,
                                       int(step))
                else:
                    step = 1
                    file_list += range(int(i.split('-', 1)[0]),
                                       int(i.split('-', 1)[1])+1,
                                       int(step))
            else:
                file_list += [int(i)]
        return list(sorted(file_list))

    def get_path(self, number):
        if number<=0:
            number = self.search_latest(number)
        fileName = f'amor{self.year}n{number:06d}.hdf'
        path = ''
        for rawd in self.rawPath:
            if os.path.exists(os.path.join(rawd, fileName)):
                path = rawd
                break
        if not path:
            from glob import glob
            potential_file = glob(f'/home/amor/data/{self.year}/*/{fileName}')
            if len(potential_file)>0:
                path = os.path.dirname(potential_file[0])
            else:
                raise FileNotFoundError(f'# ERROR: the file {fileName} can not be found '
                                        f'in {self.rawPath+["/home/amor/data"]}')
        return os.path.join(path, fileName)

    def search_latest(self, number):
        if number>0:
            raise ValueError('number needs to be relative index (negative)')
        if os.path.exists(f'/home/amor/data/{self.year}/DataNumber'):
            try:
                with open(f'/home/amor/data/{self.year}/DataNumber', 'r') as fh:
                    current_index = int(fh.readline())-1
            except Exception:
                logging.error('Can not access DataNumber', exc_info=True)
            else:
                return current_index+number
        # find all files from the given year, convert to number and return latest
        from glob import glob
        possible_files = []
        for rawd in self.rawPath:
            possible_files += glob(os.path.join(rawd, f'amor{self.year}n??????.hdf'))

        possible_files += glob(f'/home/amor/data/{self.year}/*/amor{self.year}n??????.hdf')
        possible_indices = list(set([int(os.path.basename(fi)[9:15]) for fi in possible_files]))
        possible_indices.sort()
        try:
            return possible_indices[number-1]
        except IndexError:
            raise FileNotFoundError(f'# Could not find suitable file for relative index {number} '
                                    f'in {self.rawPath+["/home/amor/data"]}, '
                                    f'possible indices {possible_indices}')




"""
Classes used to calculate projections/binnings from event data onto given grids.
"""

import logging
import warnings
from abc import ABC, abstractmethod
from typing import List, Tuple, Union

import numpy as np
from dataclasses import dataclass

from .event_data_types import EventDatasetProtocol
from .instrument import Detector, LZGrid
from .normalization import LZNormalisation

class ProjectionInterface(ABC):
    @abstractmethod
    def project(self, dataset: EventDatasetProtocol, monitor: float): ...

    @abstractmethod
    def clear(self): ...

    @abstractmethod
    def plot(self, **kwargs): ...

    @abstractmethod
    def update_plot(self): ...

@dataclass
class ProjectedReflectivity:
    R: np.ndarray
    dR: np.ndarray
    Q: np.ndarray
    dQ: np.ndarray

    @property
    def data(self):
        """
        Return combined data compatible with storing as columns in orso file.
            Q, R, dR, dQ
        """
        return np.array([self.Q, self.R, self.dR, self.dQ]).T

    def data_for_time(self, time):
        tme = np.ones(np.shape(self.Q))*time
        return np.array([self.Q, self.R, self.dR, self.dQ, tme]).T

    def scale(self, factor):
        self.R *= factor
        self.dR *= factor

    def autoscale(self, range):
        filter_q = (range[0]<=self.Q) & (self.Q<=range[1])
        filter_q &= self.dR>0
        if filter_q.sum()>0:
            scale = (self.R[filter_q]/self.dR[filter_q]).sum()/(self.R[filter_q]**2/self.dR[filter_q]).sum()
            self.scale(scale)
            logging.info(f'      scaling factor = {scale}')
            return scale
        else:
            logging.warning('      automatic scaling not possible')
            return 1.0

    def stitch(self, other: 'ProjectedReflectivity'):
        # find scaling factor between two reflectivities at points both are not zero
        filter_q = np.logical_not(np.isnan(other.R*self.R))
        filter_q &= self.R>0
        filter_q &= other.R>0
        R1 = self.R[filter_q]
        dR1 = self.dR[filter_q]
        R2 = other.R[filter_q]
        dR2 = other.dR[filter_q]
        if len(R1)>0:
            scale = (R1**2*R2**2/(dR1**2*dR2**2)).sum() / (R1**3*R2/(dR1**2*dR2**2)).sum()
            self.scale(scale)
            logging.info(f'      scaling factor = {scale}')
            return scale
        else:
            logging.warning('      automatic scaling not possible')
            return 1.0

    def subtract(self, R, dR):
        # subtract another dataset with same q-points
        self.R -= R
        self.dR = np.sqrt(self.dR**2+dR**2)

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        plt.errorbar(self.Q, self.R, yerr=self.dR, **kwargs)
        plt.yscale('log')
        plt.xlabel('Q / Å$^{-1}$')
        plt.ylabel('R')

class LZProjection(ProjectionInterface):
    grid: LZGrid
    lamda: np.ndarray
    alphaF: np.ndarray
    is_normalized: bool
    angle: float

    data: np.recarray
    _dtype = np.dtype([
            ('I', np.float64),
            ('mask', bool),
            ('ref', np.float64),
            ('err', np.float64),
            ('res', np.float64),
            ('qz', np.float64),
            ('qx', np.float64),
            ('norm', np.float64),
            ])

    def __init__(self, tthh: float, grid: LZGrid):
        self.grid = grid
        self.is_normalized = False
        self.angle = tthh

        alphaF_z  = tthh + Detector.delta_z
        lamda_l  = self.grid.lamda()
        lamda_c = (lamda_l[:-1]+lamda_l[1:])/2

        lz_shape = self.grid.lz()

        self.lamda  = lz_shape*lamda_c[:, np.newaxis]
        self.alphaF = lz_shape*alphaF_z[np.newaxis, :]
        self.data = np.zeros(self.alphaF.shape, dtype=self._dtype).view(np.recarray)
        self.data.mask = True
        self.monitor = 0.

    @classmethod
    def from_dataset(cls, dataset: EventDatasetProtocol, grid: LZGrid, has_offspecular=False):
        tthh  = dataset.geometry.nu - dataset.geometry.mu
        output = cls(tthh, grid)
        output.correct_gravity(dataset.geometry.detectorDistance)
        if has_offspecular:
            alphaI_lz = grid.lz()*(dataset.geometry.mu+dataset.geometry.kap+dataset.geometry.kad)
            output.calculate_q(alphaI_lz)
        else:
            output.calculate_q()
        return output

    def correct_gravity(self, detector_distance):
        self.alphaF += np.rad2deg( np.arctan( 3.07e-10 * detector_distance * self.lamda**2 ) )

    def calculate_q(self, alphaI=None):
        if alphaI is None:
            self.data.qz = 4.0*np.pi*np.sin(np.deg2rad(self.alphaF))/self.lamda
            self.data.qx = 0.*self.data.qz
        else:
            self.data.qz = 2.0*np.pi*(np.sin(np.deg2rad(self.alphaF))+np.sin(np.deg2rad(alphaI)))/self.lamda
            self.data.qx = 2.0*np.pi*(np.cos(np.deg2rad(self.alphaF))-np.cos(np.deg2rad(alphaI)))/self.lamda

        if self.data.qz[0,self.data.qz.shape[1]//2]  < 0:
            # assuming a 'measurement from below' when center of detector at negative qz
            self.data.qz *= -1

        self.calculate_q_resolution()

    def calculate_q_resolution(self):
        res_lz    = self.grid.lz() * 0.022**2
        res_lz    = res_lz + (0.008/self.alphaF)**2
        self.data.res    = self.data.qz * np.sqrt(res_lz)

    def apply_theta_filter(self, theta_range):
        # Filters points within theta range
        self.data.mask &= (self.alphaF<theta_range[0])|(self.alphaF>theta_range[1])

    def apply_theta_mask(self, theta_range):
        # Mask points outside theta range
        self.data.mask &= self.alphaF>=theta_range[0]
        self.data.mask &= self.alphaF<=theta_range[1]

    def apply_lamda_mask(self, lamda_range):
        # Mask points outside lambda range
        self.data.mask &= self.lamda>=lamda_range[0]
        self.data.mask &= self.lamda<=lamda_range[1]

    def apply_norm_mask(self, norm: LZNormalisation, min_norm=-1, min_theta=-1):
        # Mask points where normliazation is nan
        self.data.mask &= np.logical_not(np.isnan(norm.norm))&(norm.norm>min_norm)
        if min_theta>0:
            thsum = np.nansum(norm.norm, axis=0)
            self.data.mask &= (thsum>min_theta)[np.newaxis, :]

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        """
            Project dataset on grid and add to intensity.
            Can be called multiple times to sequentially add events.
        """
        # TODO: maybe move monitor calculation in here instead of reduction?
        e = dataset.data.events
        int_lz, *_  = np.histogram2d(e.lamda, e.detZ, bins = (self.grid.lamda(), self.grid.z()))
        self.data.I += int_lz
        self.monitor += monitor
        # in case the intensity changed one needs to normalize again
        self.is_normalized = False

    def clear(self):
        # empty data
        self.data[:] = 0
        self.data.mask = True
        self.monitor = 0.
        self.norm_monitor = 1.

    @property
    def I(self):
        output = self.data.I[:]
        output[np.logical_not(self.data.mask)] = np.nan
        return output / self.monitor

    def calc_error(self):
        # calculate error bars for resulting intensity after normalization
        self.data.err = self.data.ref * np.sqrt( 1./(self.data.I+.1) + 1./(self.data.norm+0.1) )

    def normalize_over_illuminated(self, norm: LZNormalisation):
        """
        Normalize the dataaset and take into account a difference in
        detector angle for measurement and reference.
        """
        logging.debug(f'        correcting for incident angle difference from norm {norm.angle} to data {self.angle}')
        norm_lz = norm.norm
        delta_lz = np.ones_like(norm_lz)*Detector.delta_z
        # do not perform gravity correction for footprint, would require norm detector distance that is unknown here
        fp_corr_lz = np.where(np.absolute(delta_lz+norm.angle)>5e-3,
                              (delta_lz+self.angle)/(delta_lz+norm.angle), np.nan)
        fp_corr_lz[fp_corr_lz<0] = np.nan
        self.data.mask &=  np.logical_not(np.isnan(fp_corr_lz))
        self.data.norm = norm_lz*fp_corr_lz
        self.norm_monitor = norm.monitor
        ref_lz = self.data.I/np.where(self.data.norm>0, self.data.norm, np.nan)
        ref_lz *= norm.monitor/self.monitor
        ref_lz[np.logical_not(self.data.mask)] = np.nan
        self.data.ref = ref_lz
        self.calc_error()
        self.is_normalized = True

    def normalize_no_footprint(self, norm: LZNormalisation):
        norm_lz = norm.norm
        ref_lz = (self.data.I/norm_lz)
        ref_lz *= norm.monitor/self.monitor
        ref_lz[np.logical_not(self.data.mask)] = np.nan
        self.data.norm = norm_lz
        self.data.ref = ref_lz
        self.calc_error()
        self.is_normalized = True

    def scale(self, factor: float):
        if not self.is_normalized:
            raise ValueError("Dataset needs to be normalized, first")
        self.data.ref *= factor
        self.data.err *= factor
        self.norm_monitor /= factor

    def project_on_qz(self):
        if not self.is_normalized:
            raise ValueError("Dataset needs to be normalized, first")
        q_q       = self.grid.q()
        weights_lzf = self.data.norm[self.data.mask]
        q_lzf = self.data.qz[self.data.mask]
        I_lzf = self.data.I[self.data.mask]
        dq_lzf = self.data.res[self.data.mask]

        # get number of grid points contributing to a bin, filter points with no contribution
        N_q       = np.histogram(q_lzf, bins = q_q)[0]
        N_q       = np.where(N_q > 0, N_q, np.nan)
        fltr      = N_q>0

        # calculate sum of all normalization weights per bin
        W_q       = np.maximum(np.histogram(q_lzf, bins = q_q, weights = weights_lzf)[0], 1e-10)
        # calculate sum of all dataset counts per bin
        I_q       = np.histogram(q_lzf, bins = q_q, weights = I_lzf)[0]
        # normlaize dataaset by normalization counts and scale by monitor
        R_q       = np.where(fltr, I_q*self.norm_monitor/self.monitor / W_q, np.nan)
        # error as squar-root of counts and sqrt from normalization (dR/R = sqrt( (dI/I)² + (dW/W)²)
        dR_q      = np.where(fltr, R_q*(np.sqrt(1./(I_q+0.1)+ 1./(W_q+0.1))), np.nan)
        # q-resolution is the weighted sum of individual points q-resolutions
        dq_q      = np.histogram(q_lzf, bins = q_q, weights = weights_lzf * dq_lzf )[0]
        dq_q      = np.where(fltr, dq_q/W_q, np.nan)
        return ProjectedReflectivity(R_q, dR_q, (q_q[1:]+q_q[:-1])/2., dq_q)

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        from matplotlib.colors import LogNorm

        if 'colorbar' in kwargs:
            cmap=True
            del(kwargs['colorbar'])
        else:
            cmap=False

        if self.is_normalized:
            I = self.data.ref
        else:
            I = self.data.I


        if not 'norm' in kwargs:
            vmin = I[(I>0)].min()
            vmax = np.nanmax(I)
            kwargs['norm'] = LogNorm(vmin, vmax, clip=True)


        # suppress warning for wrongly sorted y-axis pixels (blades overlap)
        with warnings.catch_warnings(action='ignore', category=UserWarning):
            self._graph = plt.pcolormesh(self.lamda, self.alphaF, I, **kwargs)
        if cmap:
            if self.is_normalized:
                plt.colorbar(label='R')
            else:
                plt.colorbar(label='I / cpm')
        plt.xlabel('$\\lambda$ / $\\AA$')
        plt.ylabel('$\\Theta$ / °')
        plt.xlim(self.lamda[0,0], self.lamda[-1,0])
        af = self.alphaF[self.data.mask]
        plt.ylim(af.min(), af.max())
        plt.title('Wavelength vs. Reflection Angle')

        self._graph_axis = plt.gca()
        plt.connect('button_press_event', self.draw_qline)

    def update_plot(self):
        """
        Inline update of previous plot by just updating the data.
        """
        from matplotlib.colors import LogNorm
        if self.is_normalized:
            I = self.data.ref
        else:
            I = self.data.I

        if isinstance(self._graph.norm, LogNorm):
            vmin = I[(I>0)].min()*0.5
        else:
            vmin = 0
        vmax = np.nanmax(I)
        self._graph.set_array(I)
        self._graph.norm.vmin = vmin
        self._graph.norm.vmax = vmax

        if self.is_normalized:
            self._graph.set_array(self.data.ref)
        else:
            self._graph.set_array(self.data.I)

    def draw_qline(self, event):
        if event.inaxes is not self._graph_axis:
            return
        from matplotlib import pyplot as plt
        tbm = self._graph_axis.figure.canvas.manager.toolbar.mode
        if event.button is plt.MouseButton.LEFT and tbm=='':
            slope = event.ydata/event.xdata
            xmax = 12.5
            self._graph_axis.plot([0, xmax], [0, slope*xmax], '-', color='grey')
            self._graph_axis.text(event.xdata, event.ydata, f'q={np.deg2rad(slope)*4.*np.pi:.3f}', backgroundcolor='white')
            plt.draw()
        if event.button is plt.MouseButton.RIGHT and tbm=='':
            for art in list(self._graph_axis.lines)+list(self._graph_axis.texts):
                art.remove()
            plt.draw()

ONLY_MAP = ['colorbar', 'cmap', 'norm']

class ReflectivityProjector(ProjectionInterface):
    lzprojection: LZProjection
    data: ProjectedReflectivity
    # TODO: maybe implement direct 1d projection in here

    def __init__(self, lzprojection, norm):
        self.lzprojection = lzprojection
        self.norm = norm

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        self.lzprojection.project(dataset, monitor)
        self.lzprojection.normalize_over_illuminated(self.norm)
        self.data = self.lzprojection.project_on_qz()

    def clear(self):
        self.lzprojection.clear()

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        for key in ONLY_MAP:
            if key in kwargs: del(kwargs[key])

        self._graph = plt.errorbar(self.data.Q, self.data.R, xerr=self.data.dQ, yerr=self.data.dR, **kwargs)
        self._graph_axis = plt.gca()
        plt.title('Reflectivity (might be improperly normalized)')
        plt.yscale('log')
        plt.xlabel('Q / $\\AA^{-1}$')
        plt.ylabel('R')

    def update_plot(self):
        ln, _, (barsx, barsy) = self._graph

        yerr_top = self.data.R+self.data.dR
        yerr_bot = self.data.R-self.data.dR
        xerr_top = self.data.Q+self.data.dQ
        xerr_bot = self.data.Q-self.data.dQ

        new_segments_x = [np.array([[xt, y], [xb, y]]) for xt, xb, y in zip(xerr_top, xerr_bot, self.data.R)]
        new_segments_y = [np.array([[x, yt], [x, yb]]) for x, yt, yb in zip(self.data.Q, yerr_top, yerr_bot)]
        barsx.set_segments(new_segments_x)
        barsy.set_segments(new_segments_y)

        ln.set_ydata(self.data.R)


class YZProjection(ProjectionInterface):
    y: np.ndarray
    z: np.ndarray

    data: np.recarray
    _dtype = np.dtype([
            ('cts', np.float64),
            ('I', np.float64),
            ('err', np.float64),
            ])

    def __init__(self):
        self.z = np.arange(Detector.nBlades*Detector.nWires+1)-0.5
        self.y = np.arange(Detector.nStripes+1)-0.5
        self.data = np.zeros((self.y.shape[0]-1, self.z.shape[0]-1), dtype=self._dtype).view(np.recarray)
        self.monitor = 0.

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        detYi, detZi, detX, delta = Detector.pixelLookUp[dataset.data.events.pixelID-1].T

        cts , *_ = np.histogram2d(detYi, detZi, bins=(self.y, self.z))
        self.data.cts += cts
        self.monitor += monitor

        self.data.I = self.data.cts / self.monitor
        self.data.err = np.sqrt(self.data.cts) / self.monitor

    def clear(self):
        self.data[:] = 0
        self.monitor = 0.

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        from matplotlib.colors import LogNorm

        if 'colorbar' in kwargs:
            cmap=True
            del(kwargs['colorbar'])
        else:
            cmap=False

        vmax = self.data.I.max()

        if not 'norm' in kwargs:
            vmin = self.data.I[(self.data.I>0)].min()*0.5
            kwargs['norm'] = LogNorm(vmin, vmax)

        self._graph = plt.pcolormesh(self.y, self.z, self.data.I.T, **kwargs)
        if cmap:
            plt.colorbar(label='I / cpm')

        plt.xlabel('Y')
        plt.ylabel('Z')
        plt.xlim(self.y[0], self.y[-1])
        plt.ylim(self.z[-1], self.z[0])
        plt.title('Horizontal Pixel vs. Vertical Pixel')

        self._graph_axis = plt.gca()
        plt.connect('button_press_event', self.draw_yzcross)

    def update_plot(self):
        """
        Inline update of previous plot by just updating the data.
        """
        from matplotlib.colors import LogNorm
        if isinstance(self._graph.norm, LogNorm):
            vmin = self.data.I[(self.data.I>0)].min()*0.5
        else:
            vmin = 0
        vmax = self.data.I.max()
        self._graph.set_array(self.data.I.T)
        self._graph.norm.vmin = vmin
        self._graph.norm.vmax = vmax

    def draw_yzcross(self, event):
        if event.inaxes is not self._graph_axis:
            return
        from matplotlib import pyplot as plt
        tbm = self._graph_axis.figure.canvas.manager.toolbar.mode
        if event.button is plt.MouseButton.LEFT and tbm=='':
            self._graph_axis.plot([event.xdata, event.xdata], [self.z[0], self.z[-1]], '-', color='grey')
            self._graph_axis.plot([self.y[0], self.y[-1]], [event.ydata, event.ydata], '-', color='grey')
            self._graph_axis.text(event.xdata, event.ydata, f'({event.xdata:.1f}, {event.ydata:.1f})', backgroundcolor='white')
            plt.draw()
        if event.button is plt.MouseButton.RIGHT and tbm=='':
            for art in list(self._graph_axis.lines)+list(self._graph_axis.texts):
                art.remove()
            plt.draw()

class YTProjection(YZProjection):
    theta: np.ndarray

    def __init__(self, tthh: float):
        dd = Detector.delta_z[1]-Detector.delta_z[0]
        delta = np.hstack([Detector.delta_z, Detector.delta_z[-1]+dd])-dd/2.
        self.theta  = tthh + delta
        super().__init__()

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        from matplotlib.colors import LogNorm

        if 'colorbar' in kwargs:
            cmap=True
            del(kwargs['colorbar'])
        else:
            cmap=False

        if not 'norm' in kwargs:
            kwargs['norm'] = LogNorm()

        self._graph = plt.pcolormesh(self.y, self.theta, self.data.I.T, **kwargs)
        if cmap:
            plt.colorbar(label='I / cpm')

        plt.xlabel('Y')
        plt.ylabel('Theta / °')
        plt.xlim(self.y[0], self.y[-1])
        plt.ylim(self.theta[-1], self.theta[0])
        plt.title('Horizontal Pixel vs. Angle')

        self._graph_axis = plt.gca()
        plt.connect('button_press_event', self.draw_tzcross)

    def draw_tzcross(self, event):
        if event.inaxes is not self._graph_axis:
            return
        from matplotlib import pyplot as plt
        tbm = self._graph_axis.figure.canvas.manager.toolbar.mode
        if event.button is plt.MouseButton.LEFT and tbm=='':
            self._graph_axis.plot([event.xdata, event.xdata], [self.theta[0], self.theta[-1]], '-', color='grey')
            self._graph_axis.plot([self.y[0], self.y[-1]], [event.ydata, event.ydata], '-', color='grey')
            self._graph_axis.text(event.xdata, event.ydata, f'({event.xdata:.1f}, {event.ydata:.1f})', backgroundcolor='white')
            plt.draw()
        if event.button is plt.MouseButton.RIGHT and tbm=='':
            for art in list(self._graph_axis.lines)+list(self._graph_axis.texts):
                art.remove()
            plt.draw()


class TofZProjection(ProjectionInterface):
    tof: np.ndarray
    z: np.ndarray

    data: np.recarray
    _dtype = np.dtype([
            ('cts', np.float64),
            ('I', np.float64),
            ('err', np.float64),
            ])

    def __init__(self, tau, foldback=False, combine=1):
        self.z = np.arange(Detector.nBlades*Detector.nWires+1)-0.5
        if foldback:
            self.tof = np.arange(0, tau, 0.0005*combine)
        else:
            self.tof = np.arange(0, 2*tau, 0.0005*combine)
        self.data = np.zeros((self.tof.shape[0]-1, self.z.shape[0]-1), dtype=self._dtype).view(np.recarray)
        self.monitor = 0.

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        detYi, detZi, detX, delta = Detector.pixelLookUp[dataset.data.events.pixelID-1].T

        cts , *_ = np.histogram2d(dataset.data.events.tof, detZi, bins=(self.tof, self.z))
        self.data.cts += cts
        self.monitor += monitor

        self.data.I = self.data.cts / self.monitor
        self.data.err = np.sqrt(self.data.cts) / self.monitor

    def clear(self):
        self.data[:] = 0
        self.monitor = 0.

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        from matplotlib.colors import LogNorm

        if 'colorbar' in kwargs:
            cmap=True
            del(kwargs['colorbar'])
        else:
            cmap=False

        if not 'norm' in kwargs:
            kwargs['norm'] = LogNorm()

        self._graph = plt.pcolormesh(self.tof*1e3, self.z, self.data.I.T, **kwargs)
        if cmap:
            plt.colorbar(label='I / cpm')

        plt.xlabel('Time of Flight / ms')
        plt.ylabel('Z')
        plt.xlim(self.tof[0]*1e3, self.tof[-1]*1e3)
        plt.ylim(self.z[-1], self.z[0])
        plt.title('Time of Flight vs. Vertical Pixel')

        self._graph_axis = plt.gca()
        plt.connect('button_press_event', self.draw_tzcross)

    def update_plot(self):
        """
        Inline update of previous plot by just updating the data.
        """
        from matplotlib.colors import LogNorm
        if isinstance(self._graph.norm, LogNorm):
            vmin = self.data.I[(self.data.I>0)].min()*0.5
        else:
            vmin = 0
        vmax = self.data.I.max()
        self._graph.set_array(self.data.I.T)
        self._graph.norm.vmin = vmin
        self._graph.norm.vmax = vmax

    def draw_tzcross(self, event):
        if event.inaxes is not self._graph_axis:
            return
        from matplotlib import pyplot as plt
        tbm = self._graph_axis.figure.canvas.manager.toolbar.mode
        if event.button is plt.MouseButton.LEFT and tbm=='':
            self._graph_axis.plot([event.xdata, event.xdata], [self.z[0], self.z[-1]], '-', color='grey')
            self._graph_axis.plot([self.tof[0]*1e3, self.tof[-1]*1e3], [event.ydata, event.ydata], '-', color='grey')
            self._graph_axis.text(event.xdata, event.ydata, f'({event.xdata:.2f}, {event.ydata:.1f})', backgroundcolor='white')
            plt.draw()
        if event.button is plt.MouseButton.RIGHT and tbm=='':
            for art in list(self._graph_axis.lines)+list(self._graph_axis.texts):
                art.remove()
            plt.draw()

class TofProjection(ProjectionInterface):
    tof: np.ndarray

    data: np.recarray
    _dtype = np.dtype([
            ('cts', np.float64),
            ('I', np.float64),
            ('err', np.float64),
            ])

    def __init__(self, tau, foldback=False):
        if foldback:
            self.tof = np.arange(0, tau, 0.0005)
        else:
            self.tof = np.arange(0, 2*tau, 0.0005)
        self.data = np.zeros(self.tof.shape[0]-1, dtype=self._dtype).view(np.recarray)
        self.monitor = 0.

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        cts , *_ = np.histogram(dataset.data.events.tof, bins=self.tof)
        self.data.cts += cts
        self.monitor += monitor

        self.data.I = self.data.cts / self.monitor
        self.data.err = np.sqrt(self.data.cts) / self.monitor

    def clear(self):
        self.data[:] = 0
        self.monitor = 0.

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        for key in ONLY_MAP:
            if key in kwargs: del(kwargs[key])


        self._graph = plt.plot(self.tof[:-1]*1e3, self.data.I, **kwargs)

        plt.xlabel('Time of Flight / ms')
        plt.ylabel('I / cpm')
        plt.xlim(self.tof[0]*1e3, self.tof[-1]*1e3)
        plt.title('Time of Flight')

    def update_plot(self):
        """
        Inline update of previous plot by just updating the data.
        """
        self._graph[0].set_ydata(self.data.I.T)

class LProjection(ProjectionInterface):
    lamda: np.ndarray

    data: np.recarray
    _dtype = np.dtype([
            ('cts', np.float64),
            ('I', np.float64),
            ('err', np.float64),
            ])

    def __init__(self):
        self.lamda = np.linspace(3.0, 12.0, 91)
        self.data = np.zeros(self.lamda.shape[0]-1, dtype=self._dtype).view(np.recarray)
        self.monitor = 0.

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        cts , *_ = np.histogram(dataset.data.events.lamda, bins=self.lamda)
        self.data.cts += cts
        self.monitor += monitor

        self.data.I = self.data.cts / self.monitor
        self.data.err = np.sqrt(self.data.cts) / self.monitor

    def clear(self):
        self.data[:] = 0
        self.monitor = 0.

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        for key in ONLY_MAP:
            if key in kwargs: del(kwargs[key])


        self._graph = plt.plot(self.lamda[:-1], self.data.I, **kwargs)

        plt.xlabel('Wavelength / Angstrom')
        plt.ylabel('I / cpm')
        plt.xlim(self.lamda[0], self.lamda[-1])
        plt.title('Wavelength')

    def update_plot(self):
        """
        Inline update of previous plot by just updating the data.
        """
        self._graph[0].set_ydata(self.data.I.T)

class TProjection(ProjectionInterface):
    theta: np.ndarray
    z: np.ndarray

    data: np.recarray
    _dtype = np.dtype([
            ('cts', np.float64),
            ('I', np.float64),
            ('err', np.float64),
            ])

    def __init__(self, tthh):
        self.z = np.arange(Detector.nBlades*Detector.nWires+1)-0.5
        dd = Detector.delta_z[1]-Detector.delta_z[0]
        delta = np.hstack([Detector.delta_z, Detector.delta_z[-1]+dd])-dd/2.
        self.theta = tthh+delta
        self.data = np.zeros(self.theta.shape[0]-1, dtype=self._dtype).view(np.recarray)
        self.monitor = 0.

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        detYi, detZi, detX, delta = Detector.pixelLookUp[dataset.data.events.pixelID-1].T

        cts , *_ = np.histogram(detZi, bins=self.z)
        self.data.cts += cts
        self.monitor += monitor

        self.data.I = self.data.cts / self.monitor
        self.data.err = np.sqrt(self.data.cts) / self.monitor

    def clear(self):
        self.data[:] = 0
        self.monitor = 0.

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        for key in ONLY_MAP:
            if key in kwargs: del(kwargs[key])


        self._graph = plt.plot(self.theta[:-1], self.data.I, **kwargs)

        plt.xlabel('Reflection Angle / °')
        plt.ylabel('I / cpm')
        plt.xlim(self.theta[-1], self.theta[0])
        plt.title('Theta')

    def update_plot(self):
        """
        Inline update of previous plot by just updating the data.
        """
        self._graph[0].set_ydata(self.data.I.T)


class CombinedProjection(ProjectionInterface):
    """
    Allows to put multiple projections together to conveniently generate combined graphs.
    """
    projections: List[ProjectionInterface]
    projection_placements: List[Union[Tuple[int, int], Tuple[int, int, int, int]]]
    grid_size: Tuple[int, int]


    def __init__(self, grid_rows, grid_cols, projections, projection_placements):
        self.projections = projections
        self.projection_placements = projection_placements
        self.grid_size = grid_rows, grid_cols

    def project(self, dataset: EventDatasetProtocol, monitor: float):
        for pi in self.projections:
            pi.project(dataset, monitor)

    def clear(self):
        for pi in self.projections:
            pi.clear()

    def plot(self, **kwargs):
        from matplotlib import pyplot as plt
        fig = plt.gcf()
        axs = fig.add_gridspec(self.grid_size[0], self.grid_size[1])
        # axs = fig.add_gridspec(self.grid_size[0]+1, self.grid_size[1],
        #                        height_ratios=[1.0 for i in range(self.grid_size[0])]+[0.2])
        self._axes = []
        for pi, placement in zip(self.projections, self.projection_placements):
            if len(placement) == 2:
                ax = fig.add_subplot(axs[placement[0], placement[1]])
            else:
                ax = fig.add_subplot(axs[placement[0]:placement[1], placement[2]:placement[3]])
            pi.plot(**dict(kwargs))
        # Create the RangeSlider
        # from matplotlib.widgets import RangeSlider
        # slider_ax = fig.add_subplot(axs[self.grid_size[0], :])
        # self._slider = RangeSlider(slider_ax, "Plot Range", 0., 1., valinit=(0., 1.))
        # self._slider.on_changed(self.update_range)

    def update_plot(self):
        for pi in self.projections:
            pi.update_plot()

    # def update_range(self, event):
    #     ...


import numba as nb
import numpy as np

@nb.jit(nb.float64[:](nb.float64[:], nb.float64, nb.float64, nb.float64),
        nopython=True, parallel=True, cache=True)
def merge_frames(tof_e, tofCut, tau, total_offset):
    # fast implementation of the merging function used in file_reader.AmorData.merge_frames
    tof_e_out = np.empty(tof_e.shape, dtype=np.float64)
    dt = (tofCut-tau)
    for ti in nb.prange(tof_e.shape[0]):
        tof_e_out[ti] = ((tof_e[ti]-dt)%tau)+total_offset  # tof shifted to 1 frame
    return tof_e_out

@nb.jit(nb.int64[:](nb.float64[:], nb.uint32[:], nb.int64[:]),
        nopython=True, parallel=True, cache=True)
def extract_walltime(tof_e, dataPacket_p, dataPacketTime_p):
    # assigning every event the wall time of the event packet (absolute time of pulse ?start?)
    totalNumber = np.shape(tof_e)[0]
    wallTime_e = np.empty(totalNumber, dtype=np.int64)
    for i in nb.prange(len(dataPacket_p)-1):
        for j in range(dataPacket_p[i], dataPacket_p[i+1]):
            wallTime_e[j] = dataPacketTime_p[i]
    for j in range(dataPacket_p[-1], totalNumber):
        wallTime_e[j] = dataPacketTime_p[-1]
    return wallTime_e

@nb.jit(nb.types.Tuple((nb.int64[:], nb.float64[:], nb.float64[:], nb.boolean[:]))
                (nb.float64[:, :], nb.uint32[:], nb.int64, nb.int64),
        nopython=True, parallel=True, cache=True)
def filter_project_x(pixelLookUp, pixelID_e, ymin, ymax):
    # project events on z-axis and create filter for events outside of y-range
    events = pixelID_e.shape[0]
    detY_e = np.empty(events, dtype=np.int64)
    detZ_e = np.empty(events, dtype=np.int64)
    detXdist_e = np.empty(events, dtype=np.float64)
    delta_e = np.empty(events, dtype=np.float64)
    mask_e = np.empty(events, dtype=nb.boolean)
    for i in nb.prange(events):
        # resolve pixel ID into y and z indicees, x position and angle
        detY_e[i] = pixelLookUp[pixelID_e[i]-1, 0]
        detZ_e[i] = pixelLookUp[pixelID_e[i]-1, 1]
        detXdist_e[i] = pixelLookUp[pixelID_e[i]-1, 2]
        delta_e[i] = pixelLookUp[pixelID_e[i]-1, 3]
        # define mask and filter y range
        mask_e[i] = (ymin<=detY_e[i]) & (detY_e[i]<=ymax)
    return (detZ_e, detXdist_e, delta_e, mask_e)

@nb.jit(nb.types.Tuple((nb.float64[:], nb.float64[:], nb.boolean[:]))
                (nb.float64[:], nb.float64[:], nb.float64[:], nb.boolean[:],
                 nb.float64, nb.float64, nb.float64, nb.float64, nb.float64, nb.float64),
        nopython=True, parallel=True, cache=True)
def calculate_derived_properties_focussing(tof_e, detXdist_e, delta_e, mask_e,
                                           lmin, lmax, nu, mu, chopperDetectorDistance, hdm):
    events = tof_e.shape[0]
    alphaF_e = np.empty(events, dtype=np.float64)
    lamda_e = np.empty(events, dtype=np.float64)
    qz_e = np.empty(events, dtype=np.float64)
    mask_e_out = np.empty(events, dtype=nb.boolean)
    denom_f1 = 1.e13*hdm
    for i in nb.prange(events):
        lamda_e[i] = denom_f1*tof_e[i]/(chopperDetectorDistance+detXdist_e[i])
        mask_e_out[i] = mask_e[i] & ((lmin<=lamda_e[i]) & (lamda_e[i]<=lmax))
        alphaF_e[i] = nu-mu+delta_e[i]
        qz_e[i] = 4*np.pi*np.sin(np.deg2rad(alphaF_e[i]))/lamda_e[i]
    return lamda_e, qz_e, mask_e



"""
Calculations performed on AmorEventData.
This module contains actions that do not need the numba base helper functions. Other actions are in event_analysis
"""
import logging
import os
import numpy as np

from .header import Header
from .options import ExperimentConfig, MonitorType
from .event_data_types import EventDatasetProtocol, EventDataAction, EVENT_BITMASKS

class ApplyPhaseOffset(EventDataAction):
    def __init__(self, chopperPhaseOffset: float):
        self.chopperPhaseOffset=chopperPhaseOffset

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        logging.debug(
            f'        replaced ch1TriggerPhase = {dataset.timing.ch1TriggerPhase} '
            f'with {self.chopperPhaseOffset}')
        dataset.timing.ch1TriggerPhase = self.chopperPhaseOffset

class ApplyParameterOverwrites(EventDataAction):
    def __init__(self, config: ExperimentConfig):
        self.config=config

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        if self.config.muOffset:
            logging.debug(f'        set muOffset = {self.config.muOffset}')
            dataset.geometry.mu += self.config.muOffset
        if self.config.mu:
            logging.debug(f'        replaced mu = {dataset.geometry.mu} with {self.config.mu}')
            dataset.geometry.mu = self.config.mu
        if self.config.nu:
            logging.debug(f'        replaced nu = {dataset.geometry.nu} with {self.config.nu}')
            dataset.geometry.nu = self.config.nu
        logging.info(f'      mu = {dataset.geometry.mu:6.3f}, '
                     f'nu = {dataset.geometry.nu:6.3f}, '
                     f'kap = {dataset.geometry.kap:6.3f}, '
                     f'kad = {dataset.geometry.kad:6.3f}')

    def update_header(self, header:Header) ->None:
        if self.config.sampleModel:
            import yaml
            from orsopy.fileio.model_language import SampleModel
            if self.config.sampleModel.endswith('.yml') or self.config.sampleModel.endswith('.yaml'):
                if os.path.isfile(self.config.sampleModel):
                    with open(self.config.sampleModel, 'r') as model_yml:
                        model = yaml.safe_load(model_yml)
                else:
                    logging.warning(f'  ! the file {self.config.sampleModel} does not exist. Ignored!')
                    return
            else:
                model = dict(stack=self.config.sampleModel)
            logging.debug(f'        set sample.model = {self.config.sampleModel}')
            header.sample.model = SampleModel.from_dict(model)


class CorrectChopperPhase(EventDataAction):
    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        dataset.data.events.tof += dataset.timing.tau*(dataset.timing.ch1TriggerPhase-dataset.timing.chopperPhase/2)/180


class CorrectSeriesTime(EventDataAction):
    def __init__(self, seriesStartTime):
        self.seriesStartTime = np.int64(seriesStartTime)

    def perform_action(self, dataset: EventDatasetProtocol)->None:
        if not 'wallTime' in dataset.data.events.dtype.names:
            raise ValueError("CorrectTimeSeries requires walltTime to be extracted, please run ExtractWalltime first")
        dataset.data.pulses.time -= self.seriesStartTime
        dataset.data.events.wallTime -= self.seriesStartTime
        dataset.data.proton_current.time -= self.seriesStartTime
        for value in dataset.data.device_logs.values():
            value.time -= self.seriesStartTime
        start, stop = dataset.data.events.wallTime[0], dataset.data.events.wallTime[-1]
        logging.debug(f'      wall time from {start/1e9:6.1f} s to {stop/1e9:6.1f} s, '
                      f'series time = {self.seriesStartTime/1e9:6.1f}')


class AssociatePulseWithMonitor(EventDataAction):
    def __init__(self, monitorType:MonitorType):
        self.monitorType = monitorType

    def perform_action(self, dataset: EventDatasetProtocol)->None:
        logging.debug(f'      using monitor type {self.monitorType}')
        if self.monitorType in [MonitorType.proton_charge or MonitorType.debug]:
            dataset.data.pulses.monitor = self.get_current_per_pulse(dataset.data.pulses.time,
                                                              dataset.data.proton_current.time,
                                                              dataset.data.proton_current.current)\
                                                              * 2*dataset.timing.tau * 1e-3
        elif self.monitorType==MonitorType.time:
            dataset.data.pulses.monitor  = 2*dataset.timing.tau
        else:  # pulses
            dataset.data.pulses.monitor  = 1

        if self.monitorType == MonitorType.debug:
            if not 'wallTime' in dataset.data.events.dtype.names:
                raise ValueError(
                    "AssociatePulseWithMonitor requires walltTime for debugging, please run ExtractWalltime first")
            cpp, t_bins = np.histogram(dataset.data.events.wallTime, dataset.data.pulses.time)
            np.savetxt('tme.hst', np.vstack((dataset.data.pulses.time[:-1], cpp, dataset.data.pulses.monitor[:-1])).T)

    @staticmethod
    def get_current_per_pulse(pulseTimeS, currentTimeS, currents):
        # add currents for early pulses and current time value after last pulse (j+1)
        currentTimeS = np.hstack([[0], currentTimeS, [pulseTimeS[-1]+1]])
        currents = np.hstack([[0], currents])
        pulseCurrentS = np.zeros(pulseTimeS.shape[0], dtype=float)
        j = 0
        for i, ti in enumerate(pulseTimeS):
            # find the last current item that was before this pulse
            while ti >= currentTimeS[j+1]:
                j += 1
            pulseCurrentS[i] = currents[j]
        return pulseCurrentS

class FilterMonitorThreshold(EventDataAction):
    def __init__(self, lowCurrentThreshold:float):
        self.lowCurrentThreshold = lowCurrentThreshold

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        if not 'wallTime' in dataset.data.events.dtype.names:
            raise ValueError(
                    "FilterMonitorThreshold requires walltTime to be extracted, please run ExtractWalltime first")
        low_current_filter = dataset.data.pulses.monitor>2*dataset.timing.tau*self.lowCurrentThreshold*1e-3
        dataset.data.pulses.monitor[np.logical_not(low_current_filter)] = 0.
        goodTimeS = dataset.data.pulses.time[low_current_filter]
        filter_e = np.logical_not(np.isin(dataset.data.events.wallTime, goodTimeS))

        dataset.data.events.mask += EVENT_BITMASKS['MonitorThreshold']*filter_e
        logging.info(f'      low-beam (<{self.lowCurrentThreshold} mC) rejected pulses: '
                     f'{dataset.data.pulses.monitor.shape[0]-goodTimeS.shape[0]} '
                     f'out of {dataset.data.pulses.monitor.shape[0]}')
        logging.info(f'          with {filter_e.sum()} events')
        if goodTimeS.shape[0]:
            logging.info(f'      average counts per pulse =  {dataset.data.events.shape[0]/goodTimeS.shape[0]:7.1f}')
        else:
            logging.info(f'      average counts per pulse = undefined')

class FilterStrangeTimes(EventDataAction):
    def perform_action(self, dataset: EventDatasetProtocol)->None:
        filter_e = np.logical_not(dataset.data.events.tof<=2*dataset.timing.tau)
        dataset.data.events.mask += EVENT_BITMASKS['StrangeTimes']*filter_e
        if filter_e.any():
            logging.warning(f'        strange times: {filter_e.sum()}')


class TofTimeCorrection(EventDataAction):
    def __init__(self, correct_chopper_opening: bool = True):
        self.correct_chopper_opening = correct_chopper_opening

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        if not 'delta' in dataset.data.events.dtype.names:
            raise ValueError(
                    "TofTimeCorrection requires delta to be extracted, please run AnalyzePixelIDs first")
        d = dataset.data
        if self.correct_chopper_opening:
            d.events.tof -= ( d.events.delta / 180. ) * dataset.timing.tau
        else:
            d.events.tof -= ( dataset.geometry.kad / 180. ) * dataset.timing.tau

class ApplyMask(EventDataAction):
    def __init__(self, bitmask_filter=None):
        self.bitmask_filter = bitmask_filter

    def perform_action(self, dataset: EventDatasetProtocol) ->None:
        # TODO: why is this action time consuming?
        d = dataset.data
        pre_filter = d.events.shape[0]
        if logging.getLogger().level <= logging.DEBUG:
            # only run this calculation if debug level is actually active
            filtered_by_mask = {}
            for key, value in EVENT_BITMASKS.items():
                filtered_by_mask[key] = ((d.events.mask & value)!=0).sum()
            logging.debug(f"        Removed by filters: {filtered_by_mask}")
        if self.bitmask_filter is None:
            d.events = d.events[d.events.mask==0]
        else:
            # remove the provided bitmask_filter bits from the events
            # this means that all bits that are set in bitmask_filter will NOT be used to filter events
            fltr = (d.events.mask & (~self.bitmask_filter)) == 0
            d.events = d.events[fltr]
        post_filter = d.events.shape[0]
        logging.info(f'      number of events: total = {pre_filter:7d}, filtered = {post_filter:7d}')
        if d.device_logs == {} or not hasattr(dataset, 'update_info_from_logs'):
            return
        # filter pulses and logs to allow update of header information
        from .helpers import add_log_to_pulses
        times = np.unique(d.events.wallTime)
        # make sure all log variables are associated with pulses
        for key, log in d.device_logs.items():
            if not key in d.pulses.dtype.names:
                # interpolate the parameter values for all existing pulses
                add_log_to_pulses(key, dataset)
        # remove all pulses that have no more events
        d.pulses = d.pulses[np.isin(d.pulses.time, times)]
        for key, log in d.device_logs.items():
            d.device_logs[key] = np.recarray(d.pulses.shape, dtype = log.dtype)
            d.device_logs[key].time = d.pulses.time
            d.device_logs[key].value = d.pulses[key]
        dataset.update_info_from_logs()


"""
Events 2 histogram, quick reduction of single file to display during experiment.
Can be used as a live preview with automatic update when files are modified.
"""

import logging
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

from orsopy import fileio
from datetime import datetime

from .file_reader import AmorEventData, AmorHeader
from .header import Header
from .instrument import LZGrid
from .normalization import LZNormalisation
from .options import E2HConfig, E2HPlotArguments, IncidentAngle, MonitorType, E2HPlotSelection
from . import event_handling as eh
from .path_handling import PathResolver
from .projection import CombinedProjection, LProjection, LZProjection, ProjectionInterface, ReflectivityProjector, \
    TofProjection,  TofZProjection,  TProjection, YTProjection, YZProjection

NEEDS_LAMDA = (E2HPlotSelection.All, E2HPlotSelection.LT, E2HPlotSelection.Q, E2HPlotSelection.L)

class E2HReduction:
    config: E2HConfig
    header: Header
    event_actions: eh.EventDataAction

    _last_mtime = 0.
    projection: ProjectionInterface

    def __init__(self, config: E2HConfig):
        self.config = config

        self.header = Header()

        self.fig = plt.figure()
        self.register_colormap()
        self.prepare_actions()

    def prepare_actions(self):
        """
        Does not do any actual reduction.
        """
        self.path_resolver = PathResolver(self.config.reader.year, self.config.reader.rawPath)
        self.file_list = self.path_resolver.resolve(self.config.reduction.fileIdentifier)
        self.file_index = 0
        self.plot_kwds = {}
        plt.rcParams.update({'font.size': self.config.reduction.fontsize})

        self.overwrite = eh.ApplyParameterOverwrites(self.config.experiment) # some actions use instrument parameters, change before that
        if self.config.reduction.update:
            # live update implies plotting
            self.config.reduction.show_plot = True

        if self.config.reduction.plot==E2HPlotSelection.Raw:
            # Raw implies fast caculations
            self.config.reduction.fast = True
        if not self.config.experiment.is_default('lambdaRange'):
            # filtering wavelength requires frame analysis
            self.config.reduction.fast = False

        if not self.config.reduction.fast or self.config.reduction.plot in NEEDS_LAMDA:
            from . import event_analysis as ea

        # Actions on datasets not used for normalization
        self.event_actions = eh.ApplyPhaseOffset(self.config.experiment.chopperPhaseOffset)
        if not self.config.reduction.fast:
            self.event_actions |= self.overwrite
            self.event_actions |= eh.CorrectChopperPhase()
            self.event_actions |= ea.ExtractWalltime()
        else:
            logging.info('    Fast reduction always uses time normalization')
            self.config.experiment.monitorType = MonitorType.time
        self.event_actions |= eh.AssociatePulseWithMonitor(self.config.experiment.monitorType)
        if self.config.experiment.monitorType in [MonitorType.proton_charge, MonitorType.debug]:
            # the filtering only makes sense if using actual monitor data, not time
            self.event_actions |= eh.FilterMonitorThreshold(self.config.experiment.lowCurrentThreshold)
        if not self.config.reduction.fast:
            self.event_actions |= eh.FilterStrangeTimes()
            if self.config.reduction.plot in [E2HPlotSelection.YT, E2HPlotSelection.YZ]:
                # perform time fold-back and apply yRange filter if not fast mode
                self.event_actions |= ea.MergeFrames()
                self.event_actions |= ea.AnalyzePixelIDs(self.config.experiment.yRange)
            if self.config.reduction.plot==E2HPlotSelection.YT:
                # perform corrections for tof if not fast mode
                self.event_actions |= eh.TofTimeCorrection(self.config.experiment.incidentAngle==IncidentAngle.alphaF)
        # select needed actions in depenence of plots
        if self.config.reduction.plot in NEEDS_LAMDA or not self.config.experiment.is_default('lambdaRange'):
            self.event_actions |= ea.MergeFrames(lamdaCut=self.config.experiment.lambdaRange[0])
            self.event_actions |= ea.AnalyzePixelIDs(self.config.experiment.yRange)
            self.event_actions |= eh.TofTimeCorrection(self.config.experiment.incidentAngle==IncidentAngle.alphaF)
            self.event_actions |= ea.CalculateWavelength(self.config.experiment.lambdaRange)
        self.event_actions |= eh.ApplyMask()

        # plot dependant options
        if self.config.reduction.plot in [E2HPlotSelection.All, E2HPlotSelection.LT, E2HPlotSelection.Q]:
            self.grid = LZGrid(0.05, [0.0, 0.25], lambda_overwrite=self.config.experiment.lambdaRange)
            self.grid.dldl = 0.01

        if self.config.reduction.plot in [E2HPlotSelection.All, E2HPlotSelection.Raw,
                                          E2HPlotSelection.LT, E2HPlotSelection.YT,
                                          E2HPlotSelection.YZ, E2HPlotSelection.TZ]:
            self.plot_kwds['colorbar'] = True
            self.plot_kwds['cmap'] = str(self.config.reduction.plot_colormap)
            if self.config.reduction.plotArgs==E2HPlotArguments.Linear:
                self.plot_kwds['norm'] = None

    def reduce(self):
        if self.config.reduction.plot in [E2HPlotSelection.All, E2HPlotSelection.LT, E2HPlotSelection.Q]:
            if self.config.reduction.normalizationModel:
                self.norm = LZNormalisation.model(self.grid)
            else:
                self.norm = LZNormalisation.unity(self.grid)

        self.prepare_graphs()

        while self.file_index < len(self.file_list):
            self.read_data()
            self.add_data()

        if self.config.reduction.plotArgs==E2HPlotArguments.OutputFile:
            self.create_file_output()
        if self.config.reduction.plotArgs!=E2HPlotArguments.OutputFile or self.config.reduction.show_plot:
            self.create_graph()

        if self.config.reduction.plotArgs==E2HPlotArguments.Default and not self.config.reduction.update:
            # safe to image file if not auto-updating graph
            plt.savefig(f'e2h_{self.config.reduction.plot}.png', dpi=300)
        if self.config.reduction.kafka:
            from .kafka_serializer import ESSSerializer
            self.serializer = ESSSerializer()
            self.fig.canvas.mpl_connect('close_event', self.serializer.end_command_thread)
            self.serializer.start_command_thread()
            self.serializer.send(self.projection)
        if self.config.reduction.update:
            self.timer = self.fig.canvas.new_timer(1000)
            self.timer.add_callback(self.update)
            self.timer.start()
        if self.config.reduction.show_plot:
            plt.show()


    def register_colormap(self):
        cmap = plt.colormaps['turbo'](np.arange(256))
        cmap[:1, :] = np.array([256/256, 255/256, 236/256, 1])
        cmap = ListedColormap(cmap, name='jochen_deluxe', N=cmap.shape[0])
        #cmap.set_bad((1.,1.,0.9))
        plt.colormaps.register(cmap)

    def prepare_graphs(self):
        last_file_header = AmorHeader(self.file_list[-1])
        self.overwrite.perform_action(last_file_header)
        tthh  = last_file_header.geometry.nu - last_file_header.geometry.mu

        if not self.config.reduction.is_default('thetaRangeR'):
            # adjust range based on detector center
            thetaRange = [ti+tthh for ti in self.config.reduction.thetaRangeR]
        else:
            thetaRange = [tthh - last_file_header.geometry.div/2, tthh + last_file_header.geometry.div/2]

        if self.config.reduction.plot==E2HPlotSelection.LT:
            self.projection = LZProjection(tthh, self.grid)
            if not self.config.reduction.fast:
                self.projection.correct_gravity(last_file_header.geometry.detectorDistance)
            self.projection.apply_lamda_mask(self.config.experiment.lambdaRange)
            self.projection.apply_theta_mask(thetaRange)
            for thi in self.config.reduction.thetaFilters:
                self.projection.apply_theta_filter((thi[0]+tthh, thi[1]+tthh))
            self.projection.apply_norm_mask(self.norm)

        if self.config.reduction.plot==E2HPlotSelection.Q:
            plz = LZProjection(tthh, self.grid)
            if not self.config.reduction.fast:
                plz.correct_gravity(last_file_header.geometry.detectorDistance)
            plz.calculate_q()
            plz.apply_lamda_mask(self.config.experiment.lambdaRange)
            plz.apply_theta_mask(thetaRange)
            for thi in self.config.reduction.thetaFilters:
                self.projection.apply_theta_filter((thi[0]+tthh, thi[1]+tthh))
            plz.apply_norm_mask(self.norm)
            self.projection = ReflectivityProjector(plz, self.norm)

        if self.config.reduction.plot==E2HPlotSelection.YZ:
            self.projection = YZProjection()

        if self.config.reduction.plot==E2HPlotSelection.YT:
            self.projection = YTProjection(tthh)

        if self.config.reduction.plot==E2HPlotSelection.T:
            self.projection = TProjection(tthh)

        if self.config.reduction.plot==E2HPlotSelection.L:
            self.projection = LProjection()

        if self.config.reduction.plot==E2HPlotSelection.TZ:
            self.projection = TofZProjection(last_file_header.timing.tau, foldback=not self.config.reduction.fast)

        if self.config.reduction.plot==E2HPlotSelection.ToF:
            self.projection = TofProjection(last_file_header.timing.tau, foldback=not self.config.reduction.fast)

        if self.config.reduction.plot==E2HPlotSelection.All:
            plz = LZProjection(tthh, self.grid)
            if not self.config.reduction.fast:
                plz.correct_gravity(last_file_header.geometry.detectorDistance)
            plz.calculate_q()
            plz.apply_lamda_mask(self.config.experiment.lambdaRange)
            plz.apply_theta_mask(thetaRange)
            for thi in self.config.reduction.thetaFilters:
                plz.apply_theta_filter((thi[0]+tthh, thi[1]+tthh))
            plz.apply_norm_mask(self.norm)
            pr = ReflectivityProjector(plz, self.norm)
            pyz = YZProjection()
            self.projection = CombinedProjection(3, 2, [plz, pyz, pr],
                                                 [(0, 2, 0, 1), (0, 2, 1, 2), (2, 3, 0, 2)])

        if self.config.reduction.plot==E2HPlotSelection.Raw:
            del(self.plot_kwds['colorbar'])
            # A combined graph that does not require longer calculations
            plyt = YTProjection(tthh)
            pltofz = TofZProjection(last_file_header.timing.tau, foldback=not self.config.reduction.fast)
            pltof =  TofProjection(last_file_header.timing.tau, foldback=not self.config.reduction.fast)
            plt = TProjection(tthh)

            self.projection = CombinedProjection(3, 3, [plyt, pltofz, plt, pltof],
                                                 [(0,2, 0, 1), (0, 2, 1, 3), (2,3, 0,1),(2,3,1,3)])

    def read_data(self):
        fileName = self.file_list[self.file_index]
        self.dataset = AmorEventData(fileName, max_events=self.config.reduction.max_events)
        if self.dataset.EOF or fileName==self.file_list[-1]:
            self.file_index += 1
        self.event_actions(self.dataset)
        self.dataset.update_header(self.header)

        self.header.measurement_data_files.append(fileio.File(file=os.path.basename(fileName),
                                                              timestamp=self.dataset.fileDate))

    def add_data(self):
        self.monitor = self.dataset.data.pulses.monitor.sum()
        self.projection.project(self.dataset, monitor=self.monitor)
        if self.config.reduction.plot==E2HPlotSelection.LT:
            self.projection.normalize_over_illuminated(self.norm)

    def create_file_output(self):
        raise NotImplementedError("Export to text output not yet implemented")

    def create_title(self):
        output = "Events to Histogram - "
        output += ",".join(["#"+os.path.basename(fi)[9:15].lstrip('0') for fi in self.file_list])
        output += f" ($\\mu$={self.dataset.geometry.mu:.2f} ;"
        output += f" $\\nu$={self.dataset.geometry.nu:.2f})"
        if self.config.reduction.update:
            output += f"\n at "+datetime.now().strftime("%m/%d/%Y %H:%M:%S")
        return output

    def create_graph(self):
        plt.suptitle(self.create_title())
        self.projection.plot(**self.plot_kwds)
        plt.tight_layout(pad=0.5)

    def replace_dataset(self, latest):
        new_files = self.path_resolver.resolve(f'{latest}')
        if not os.path.exists(new_files[-1]):
            return
        try:
            # check that events exist in the new file
            AmorEventData(new_files[-1], 0, max_events=1_000)
        except Exception:
            logging.debug("Problem when trying to load new dataset", exc_info=True)
            return

        logging.warning(f"Preceding to next file {latest}")
        self.file_list = new_files
        self.file_index = 0
        self.prepare_actions()
        self.prepare_graphs()
        self.read_data()
        self.add_data()
        self.fig.clear()
        self.create_graph()
        plt.draw()

    def update(self):
        logging.debug("    check for update")
        if self.config.reduction.fileIdentifier=='0':
            # if latest file was choosen, check if new one available and switch to it
            current = int(os.path.basename(self.file_list[-1])[9:15])
            latest = self.path_resolver.search_latest(0)
            if latest>current:
                self.replace_dataset(latest)
                return
        # if all events were read last time, only load more if file was modified
        if self.dataset.EOF and os.path.getmtime(self.file_list[-1])<=self._last_mtime:
            return

        self._last_mtime = os.path.getmtime(self.file_list[-1])
        try:
            update_data = AmorEventData(self.file_list[-1], self.dataset.last_index+1,
                                        max_events=self.config.reduction.max_events)
        except EOFError:
            return
        logging.info("    updating with new data")

        self.event_actions(update_data)
        self.dataset=update_data
        self.monitor = self.dataset.data.pulses.monitor.sum()
        self.projection.project(update_data, self.monitor)

        self.projection.update_plot()
        plt.suptitle(self.create_title())
        plt.draw()

        if self.config.reduction.kafka:
            self.serializer.send(self.projection)



import logging
import os
import sys

import numpy as np
from orsopy import fileio

from .event_analysis import FilterByLog
from .event_handling import ApplyMask
from .file_reader import AmorEventData
from .header import Header
from .path_handling import PathResolver
from .options import ReflectivityConfig, IncidentAngle, MonitorType, NormalisationMethod, MONITOR_UNITS
from .instrument import LZGrid
from .normalization import LZNormalisation
from . import event_handling as eh, event_analysis as ea
from .projection import LZProjection


class ReflectivityReduction:
    config: ReflectivityConfig
    header: Header
    normevent_actions: eh.EventDataAction
    dataevent_actions: eh.EventDataAction
    
    def __init__(self, config: ReflectivityConfig):
        self.config = config

        self.header = Header()
        self.header.reduction.call = config.call_string()

        self.prepare_actions()

    def prepare_actions(self):
        """
        Prepare the actions applied to each event dataset, does not do any actual reduction.
        """
        self.path_resolver = PathResolver(self.config.reader.year, self.config.reader.rawPath)

        # setup all actions performed on event datasets before projection on the grid
        # The order of these corrections matter as some rely on parameters modified before
        if self.config.reduction.normalisationFileIdentifier:
            # explicit steps performed on AmorEventDataset for normalization files
            self.normevent_actions = eh.ApplyPhaseOffset(self.config.experiment.chopperPhaseOffset)
            self.normevent_actions |= eh.CorrectChopperPhase()
            self.normevent_actions |= eh.AssociatePulseWithMonitor(self.config.experiment.monitorType)
            if self.config.experiment.monitorType in [MonitorType.proton_charge, MonitorType.debug]:
                self.normevent_actions |= ea.ExtractWalltime()
                self.normevent_actions |= eh.FilterMonitorThreshold(self.config.experiment.lowCurrentThreshold)
            self.normevent_actions |= eh.FilterStrangeTimes()
            self.normevent_actions |= ea.MergeFrames()
            self.normevent_actions |= ea.AnalyzePixelIDs(self.config.experiment.yRange)
            self.normevent_actions |= eh.TofTimeCorrection(self.config.experiment.incidentAngle==IncidentAngle.alphaF)
            self.normevent_actions |= ea.CalculateWavelength(self.config.experiment.lambdaRange)
            self.normevent_actions |= eh.ApplyMask()
        # Actions on datasets not used for normalization
        self.dataevent_actions = eh.ApplyPhaseOffset(self.config.experiment.chopperPhaseOffset)
        self.dataevent_actions |= eh.ApplyParameterOverwrites(self.config.experiment) # some actions use instrument parameters, change before that
        self.dataevent_actions |= eh.CorrectChopperPhase()
        self.dataevent_actions |= ea.ExtractWalltime()
        self.dataevent_time_correction = eh.CorrectSeriesTime(0) # will be set from first dataset
        self.dataevent_actions |= self.dataevent_time_correction
        self.dataevent_actions |= eh.AssociatePulseWithMonitor(self.config.experiment.monitorType)
        if self.config.experiment.monitorType in [MonitorType.proton_charge or MonitorType.debug]:
            # the filtering only makes sense if using actual monitor data, not time
            self.dataevent_actions |= eh.FilterMonitorThreshold(self.config.experiment.lowCurrentThreshold)
        self.dataevent_actions |= eh.FilterStrangeTimes()
        self.dataevent_actions |= ea.MergeFrames()
        self.dataevent_actions |= ea.AnalyzePixelIDs(self.config.experiment.yRange)
        self.dataevent_actions |= eh.TofTimeCorrection(self.config.experiment.incidentAngle==IncidentAngle.alphaF)
        self.dataevent_actions |= ea.CalculateWavelength(self.config.experiment.lambdaRange)
        self.dataevent_actions |= ea.CalculateQ(self.config.experiment.incidentAngle)
        if not self.config.reduction.is_default('qzRange'):
            self.dataevent_actions |= ea.FilterQzRange(self.config.reduction.qzRange)
        for lf in self.config.reduction.logfilter:
            self.dataevent_actions |= ea.FilterByLog(lf)
        self.dataevent_actions |= eh.ApplyMask()

        self.grid = LZGrid(self.config.reduction.qResolution, self.config.reduction.qzRange)

    def reduce(self):
        if not os.path.exists(f'{self.config.output.outputPath}'):
            logging.debug(f'Creating destination path {self.config.output.outputPath}')
            os.system(f'mkdir {self.config.output.outputPath}')

        # load or create normalisation matrix
        if self.config.reduction.normalisationFileIdentifier:
            # TODO: change option definition to single normalization short_code
            self.create_normalisation_map(self.config.reduction.normalisationFileIdentifier[0])
        else:
            self.norm = LZNormalisation.unity(self.grid)

        if self.config.reduction.normalizationSmoothing:
            self.norm.smooth(self.config.reduction.normalizationSmoothing)

        # load R(q_z) curve to be subtracted:
        if self.config.reduction.subtract:
            self.sq_q, self.sR_q, self.sdR_q, self.sFileName = self.loadRqz(self.config.reduction.subtract)
            logging.warning(f'loaded background file: {self.sFileName}')
            self.header.reduction.corrections.append(f'background from \'{self.sFileName}\' subtracted')
            self.subtract = True
        else:
            self.subtract = False

        # load measurement data and do the reduction
        self.datasetsRqz = []
        self.datasetsRlt = []
        for i, short_notation in enumerate(self.config.reduction.fileIdentifier):
            self.read_file_block(i, short_notation)

        # output
        if self.config.output.is_default('outputName'):
            import datetime
            _date = datetime.datetime.now().replace(microsecond=0).isoformat().replace(':', '-')
            if self.header.sample.name:
                _sampleName = self.header.sample.name.replace(' ', '_')
            else:
                _sampleName = 'unknown'
            _mu = int(self.dataset.geometry.mu * 3)
            self.config.output.outputName = f'{_sampleName}_{_mu:03}_{_date}'

        logging.warning('output:')

        if 'Rqz.ort' in self.config.output.outputFormats:
            self.save_Rqz()

        if 'Rlt.ort' in self.config.output.outputFormats:
            self.save_Rtl()

        if self.config.output.plot:
            import matplotlib.pyplot as plt
            if 'Rqz.ort' in self.config.output.outputFormats:
                plt.figure(num=99)
                plt.legend()
            plt.show()

    def read_file_block(self, i, short_notation):
        logging.warning('input:')
        file_list = self.path_resolver.resolve(short_notation)

        self.header.measurement_data_files = []

        self.dataset = AmorEventData(file_list[0])
        if self.config.experiment.monitorType==MonitorType.auto:
            if self.dataset.data.proton_current.current.sum()>1:
                self.config.experiment.monitorType = MonitorType.proton_charge
                logging.debug('      monitor type set to "proton current"')
            else:
                self.config.experiment.monitorType = MonitorType.time
                logging.debug('      monitor type set to "time"')
            # update actions to sue selected monitor
            self.prepare_actions()
            # reload normalization to make sure the monitor matches
            if self.config.reduction.normalisationFileIdentifier:
                self.create_normalisation_map(self.config.reduction.normalisationFileIdentifier[0])

        self.dataevent_time_correction.seriesStartTime = self.dataset.eventStartTime
        self.dataevent_actions(self.dataset)
        self.dataset.update_header(self.header)
        self.dataevent_actions.update_header(self.header)
        for fi in file_list[1:]:
            di = AmorEventData(fi)
            self.dataevent_actions(di)
            self.dataset.append(di)

        for fileName in file_list:
            self.header.measurement_data_files.append(fileio.File( file=os.path.basename(fileName),
                                                                   timestamp=self.dataset.fileDate))

        if 'polarization_config_label' in self.dataset.data.device_logs:
            pols = np.unique(self.dataset.data.device_logs['polarization_config_label'].value)
            pols = pols[pols>0]
            if len(pols)>1:
                logging.warning(f'    found {len(pols)} polarization configurations, splitting dataset accordingly')
                from copy import deepcopy
                from . import const
                full_ds = deepcopy(self.dataset)
                for pi in pols:
                    plabel = const.polarizationLabels[pi]
                    pol_filter = FilterByLog(f'polarization_config_label=={pi}',
                                             remove_switchpulse=True) | ApplyMask()
                    logging.info(f'    filter {plabel} using polarization_config_label=={pi}')
                    pol_filter(self.dataset)
                    self.dataset.update_header(self.header)
                    pol_filter.update_header(self.header)
                    if self.config.reduction.timeSlize:
                        if i>0:
                            logging.warning(
                                "    time slizing should only be used for one set of datafiles, check parameters")
                        self.analyze_timeslices(i, polstr=f' : polarization = {plabel}')
                    else:
                        self.analyze_unsliced(i, polstr=f' : polarization = {plabel}')
                    self.dataset = deepcopy(full_ds)
                return
        if self.config.reduction.timeSlize:
            if i>0:
                logging.warning("    time slizing should only be used for one set of datafiles, check parameters")
            self.analyze_timeslices(i)
        else:
            self.analyze_unsliced(i)

    def analyze_unsliced(self, i, polstr=''):
        self.monitor = self.dataset.data.pulses.monitor.sum()
        logging.info(f'    monitor = {self.monitor:8.2f} {MONITOR_UNITS[self.config.experiment.monitorType]}')

        proj:LZProjection = self.project_on_lz()
        try:
            scale = self.config.reduction.scale[i]
        except IndexError:
            scale = self.config.reduction.scale[-1]
        proj.scale(scale)

        if 'Rqz.ort' in self.config.output.outputFormats:
            headerRqz = self.header.orso_header()
            headerRqz.data_set = f'Nr {i} : mu = {self.dataset.geometry.mu:6.3f} deg{polstr}'

            # projection on qz-grid
            result = proj.project_on_qz()

            if self.config.reduction.autoscale:
                if i==0:
                    result.autoscale(self.config.reduction.autoscale)
                else:
                    result.stitch(self.last_result)

            if self.subtract:
                if len(result.Q)==len(self.sq_q):
                    result.subtract(self.sR_q, self.sdR_q)
                else:
                    logging.warning(
                            f'backgroung file {self.sFileName} not compatible with q_z scale ({len(self.sq_q)} vs. {len(result.Q)})')

            orso_data = fileio.OrsoDataset(headerRqz, result.data)
            self.last_result = result
            self.datasetsRqz.append(orso_data)

            if self.config.output.plot:
                import matplotlib.pyplot as plt
                # plot all reflectivity results in same graph
                plt.figure(num=99)
                result.plot(label=f'{self.config.reduction.fileIdentifier[i]}')
        if 'Rlt.ort' in self.config.output.outputFormats:
            columns = [
                fileio.Column('Qz', '1/angstrom', 'normal momentum transfer'),
                fileio.Column('R', '', 'specular reflectivity'),
                fileio.ErrorColumn(error_of='R', error_type='uncertainty', value_is='sigma'),
                fileio.ErrorColumn(error_of='Qz', error_type='resolution', value_is='sigma'),
                fileio.Column('lambda', 'angstrom', 'wavelength'),
                fileio.Column('alpha_f', 'deg', 'final angle'),
                fileio.Column('l', '', 'index of lambda-bin'),
                fileio.Column('t', '', 'index of theta bin'),
                fileio.Column('intensity', '', 'filtered neutron events per pixel'),
                fileio.Column('norm', '', 'normalisation matrix'),
                fileio.Column('mask', '', 'pixels used for calculating R(q_z)'),
                fileio.Column('Qx', '1/angstrom', 'parallel momentum transfer'),
                ]

            ts, zs = proj.data.shape
            lindex_lz = np.tile(np.arange(1, ts+1), (zs, 1)).T
            tindex_lz = np.tile(np.arange(1, zs+1), (ts, 1))

            j = 0
            for item in zip(
                    proj.data.qz.T,
                    proj.data.ref.T,
                    proj.data.err.T,
                    proj.data.res.T,
                    proj.lamda.T,
                    proj.alphaF.T,
                    lindex_lz.T,
                    tindex_lz.T,
                    proj.data.I.T,
                    proj.data.norm.T,
                    proj.data.mask.T,
                    proj.data.qx.T,
                    ):
                data = np.array(list(item)).T
                headerRlt = self.header.orso_header(columns=columns)
                headerRlt.data_set = f'dataset_{i}_{j+1} : alpha_f = {proj.alphaF[0, j]:6.3f} deg'
                orso_data = fileio.OrsoDataset(headerRlt, data)
                self.datasetsRlt.append(orso_data)
                j += 1

            if self.config.output.plot:
                import matplotlib.pyplot as plt
                plt.figure()
                proj.plot(colorbar=True, cmap=str(self.config.output.plot_colormap))
                plt.title(f'{self.config.reduction.fileIdentifier[i]}')

    def analyze_timeslices(self, i, polstr=''):
        wallTime_e = np.float64(self.dataset.data.events.wallTime)/1e9
        pulseTimeS = np.float64(self.dataset.data.pulses.time)/1e9
        interval = self.config.reduction.timeSlize[0]
        try:
            start = self.config.reduction.timeSlize[1]
        except IndexError:
            start = 0
        try:
            stop = self.config.reduction.timeSlize[2]
        except IndexError:
            stop = wallTime_e[-1]
        # make overwriting log lines possible by removing newline at the end
        #logging.StreamHandler.terminator = "\r"
        logging.warning(f'    time slizing')
        logging.info('      slize  time  monitor')
        for ti, time in enumerate(np.arange(start, stop, interval)):
            slice = self.dataset.get_timeslice(time, time+interval)
            self.monitor = np.sum(slice.data.pulses.monitor)
            logging.info(f'      {ti:<4d}  {time:6.0f}  {self.monitor:7.2f} {MONITOR_UNITS[self.config.experiment.monitorType]}')

            proj: LZProjection = self.project_on_lz(slice)
            try:
                scale = self.config.reduction.scale[i]
            except IndexError:
                scale = self.config.reduction.scale[-1]
            proj.scale(scale)

            # projection on qz-grid
            result = proj.project_on_qz()

            if self.config.reduction.autoscale:
                # scale every slice the same
                if ti==0:
                    if i==0:
                        atscale = result.autoscale(self.config.reduction.autoscale)
                    else:
                        atscale = result.stitch(self.last_result)
                else:
                    result.scale(atscale)

            if self.subtract:
                if len(result.Q)==len(self.sq_q):
                    result.subtract(self.sR_q, self.sdR_q)
                else:
                    logging.warning(
                            f'backgroung file {self.sFileName} not compatible with q_z scale ({len(self.sq_q)} vs. {len(result.Q)})')

            headerRqz = self.header.orso_header(
                    extra_columns=[fileio.Column('time', 's', 'time relative to start of measurement series')])
            headerRqz.data_set = f'{i}_{ti}: time = {time:8.1f} s  to {time+interval:8.1f} s{polstr}'
            orso_data = fileio.OrsoDataset(headerRqz, result.data_for_time(time))
            self.datasetsRqz.append(orso_data)

            if self.config.output.plot:
                import matplotlib.pyplot as plt
                # plot all reflectivity results in same graph
                plt.figure(num=99)
                result.plot(label=f'{self.config.reduction.fileIdentifier[i]} @ {time:.1f}s')

        self.last_result = result
        # reset normal logging behavior
        #logging.StreamHandler.terminator = "\n"
        logging.info(f'      done  {min(time+interval, pulseTimeS[-1]):5.0f}')

    def save_Rqz(self):
        fname = os.path.join(self.config.output.outputPath, f'{self.config.output.outputName}.Rqz.ort')
        logging.warning(f'    {fname}')
        if os.path.exists(fname) and self.config.output.append:
            logging.info('        file already exists, append as new dataset')
            with open(fname, 'r') as f:
                f.readline()
                theSecondLine = f.readline()[3:]
            prev_data = fileio.load_orso(fname)
            prev_names = [di.info.data_set for di in prev_data]
            for i, di in enumerate(self.datasetsRqz):
                while di.info.data_set in prev_names:
                    if di.info.data_set.startswith('Nr '):
                        di.info.data_set = f'Nr {i+len(prev_data)} :'+di.info.data_set.split(':', 1)[1]
                        break
                    di.info.data_set = di.info.data_set+'_'
            fileio.save_orso(prev_data+self.datasetsRqz, fname, data_separator='\n', comment=theSecondLine)
        else:
            theSecondLine = f' {self.header.experiment.title} | {self.header.experiment.start_date} | sample {self.header.sample.name} | R(q_z)'
            fileio.save_orso(self.datasetsRqz, fname, data_separator='\n', comment=theSecondLine)

    def save_Rtl(self):
        fname = os.path.join(self.config.output.outputPath, f'{self.config.output.outputName}.Rlt.ort')
        logging.warning(f'    {fname}')
        theSecondLine = f' {self.header.experiment.title} | {self.header.experiment.start_date} | sample {self.header.sample.name} | R(lambda, theta)'
        fileio.save_orso(self.datasetsRlt, fname, data_separator='\n', comment=theSecondLine)

    def loadRqz(self, name):
        fname = os.path.join(self.config.output.outputPath, name)
        if os.path.exists(fname):
            fileName = fname
        elif os.path.exists(f'{fname}.Rqz.ort'):
            fileName = f'{fname}.Rqz.ort'
        else:
            sys.exit(f'### the background file \'{fname}\' does not exist! => stopping')

        q_q, Sq_q, dS_q = np.loadtxt(fileName, usecols=(0, 1, 2), comments='#', unpack=True)

        return q_q, Sq_q, dS_q, fileName

    def create_normalisation_map(self, short_notation):
        outputPath = self.config.output.outputPath
        normalisation_list = self.path_resolver.expand_file_list(short_notation)
        name = '_'.join(map(str, normalisation_list))
        n_path = os.path.join(outputPath, f'{name}.norm')

        self.norm = None
        if os.path.exists(n_path):
            logging.debug(f'trying to load matrix from file {n_path}')
            try:
                self.norm = LZNormalisation.from_file(n_path, self.normevent_actions.action_hash())
            except (ValueError, EOFError):
                self.norm =None
            else:
                logging.warning(f'normalisation matrix: found and using {n_path}')
        if self.norm is None:
            # in case file does not exist or the action hash doesn't match, create new normalization
            logging.warning(f'normalisation matrix: using the files {normalisation_list}')
            normalization_files = list(map(self.path_resolver.get_path, normalisation_list))
            reference = AmorEventData(normalization_files[0])
            self.normevent_actions(reference)
            for nfi in normalization_files[1:]:
                toadd = AmorEventData(nfi)
                self.normevent_actions(toadd)
                reference.append(toadd)
            self.norm = LZNormalisation(reference, self.config.reduction.normalisationMethod, self.grid)
            if reference.data.events.shape[0] > 1e6:
                self.norm.safe(n_path, self.normevent_actions.action_hash())

        self.norm.update_header(self.header)
        self.header.reduction.corrections.append('normalisation with \'additional files\'')

    def project_on_lz(self, dataset=None):
        if dataset is None:
            dataset=self.dataset
        proj = LZProjection.from_dataset(dataset, self.grid,
                                         has_offspecular=(self.config.experiment.incidentAngle!=IncidentAngle.alphaF))

        t0 = dataset.geometry.nu-dataset.geometry.mu
        if not self.config.reduction.is_default('thetaRangeR'):
            # adjust range based on detector center
            thetaRange = [ti+t0 for ti in self.config.reduction.thetaRangeR]
            proj.apply_theta_mask(thetaRange)
        elif not self.config.reduction.is_default('thetaRange'):
            proj.apply_theta_mask(self.config.reduction.thetaRange)
        else:
            thetaRange = [dataset.geometry.nu - dataset.geometry.mu - dataset.geometry.div/2,
                                              dataset.geometry.nu - dataset.geometry.mu + dataset.geometry.div/2]
            proj.apply_theta_mask(thetaRange)
        for thi in self.config.reduction.thetaFilters:
            # apply theta filters relative to angle on detector (issues with parts of the incoming divergence)
            proj.apply_theta_filter((thi[0]+t0, thi[1]+t0))

        proj.apply_lamda_mask(self.config.experiment.lambdaRange)

        proj.apply_norm_mask(self.norm, min_norm=self.config.reduction.normalizationFilter,
                             min_theta=self.config.reduction.normAngleFilter)

        proj.project(dataset, self.monitor)

        if self.config.reduction.normalisationMethod == NormalisationMethod.over_illuminated:
            logging.debug('      assuming an overilluminated sample and correcting for the angle of incidence')
            proj.normalize_over_illuminated(self.norm)
        elif self.config.reduction.normalisationMethod==NormalisationMethod.under_illuminated:
            logging.debug('      assuming an underilluminated sample and ignoring the angle of incidence')
            proj.normalize_no_footprint(self.norm)
        elif self.config.reduction.normalisationMethod==NormalisationMethod.direct_beam:
            logging.debug('      assuming direct beam for normalisation and ignoring the angle of incidence')
            proj.normalize_no_footprint(self.norm)
        else:
            logging.error('unknown normalisation method! Use [u]nder, [o]ver or [d]irect illumination')
            proj.normalize_no_footprint(self.norm)
        if self.monitor<=1e-6:
            logging.info('                               low monitor -> nan output')
            proj.data.ref *= np.nan

        return proj



import os
import cProfile
import numpy as np
from unittest import TestCase
from dataclasses import fields, MISSING
from eos import options, reduction_reflectivity, logconfig
from orsopy import fileio

logconfig.setup_logging()
logconfig.update_loglevel(1)

# TODO: add unit tests for individual parts of reduction

class FullAmorTest(TestCase):
    @classmethod
    def setUpClass(cls):
        # generate map for option defaults
        cls._field_defaults = {}
        for opt in [options.ExperimentConfig, options.ReflectivityReductionConfig, options.ReflectivityOutputConfig]:
            defaults = {}
            for field in fields(opt):
                if field.default not in [None, MISSING]:
                    defaults[field.name] = field.default
                elif field.default_factory not in [None, MISSING]:
                    defaults[field.name] = field.default_factory()
            cls._field_defaults[opt.__name__] = defaults
        cls.pr = cProfile.Profile()

    @classmethod
    def tearDownClass(cls):
        cls.pr.dump_stats("profile_test.prof")

    def setUp(self):
        self.pr.enable()
        self.reader_config = options.ReaderConfig(
                year=2025,
                rawPath=["test_data"],
                )

    def tearDown(self):
        self.pr.disable()
        for fi in ['test_results/test.Rqz.ort', 'test_results/5952.norm']:
           try:
               os.unlink(fi)
           except FileNotFoundError:
               pass

    def test_time_slicing(self):
        experiment_config = options.ExperimentConfig(
                chopperPhase=-13.5,
                chopperPhaseOffset=-5,
                yRange=(18, 48),
                lambdaRange=(3., 11.5),
                mu=0,
                nu=0,
                muOffset=0.0,
                sampleModel='air | 10 H2O | D2O'
                )
        reduction_config = options.ReflectivityReductionConfig(
                qResolution=0.01,
                qzRange=(0.01, 0.15),
                thetaRange=(-0.75, 0.75),
                fileIdentifier=["6003-6005"],
                scale=[1],
                normalisationFileIdentifier=[],
                timeSlize=[300.0]
                )
        output_config = options.ReflectivityOutputConfig(
                outputFormats=[options.OutputFomatOption.Rqz_ort],
                outputName='test',
                outputPath='test_results',
                )
        config=options.ReflectivityConfig(self.reader_config, experiment_config, reduction_config, output_config)
        # run three times to get similar timing to noslicing runs
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()

    def test_noslicing(self):
        experiment_config = options.ExperimentConfig(
                chopperPhase=-13.5,
                chopperPhaseOffset=-5,
                yRange=(18, 48),
                lambdaRange=(3., 11.5),
                mu=0,
                nu=0,
                muOffset=0.0,
                )
        reduction_config = options.ReflectivityReductionConfig(
                qResolution=0.01,
                thetaRange=(-0.75, 0.75),
                fileIdentifier=["6003", "6004", "6005"],
                scale=[1],
                normalisationFileIdentifier=["5952"],
                autoscale=(0.0, 0.05),
                )
        output_config = options.ReflectivityOutputConfig(
                outputFormats=[options.OutputFomatOption.Rqz_ort],
                outputName='test',
                outputPath='test_results',
                )
        config=options.ReflectivityConfig(self.reader_config, experiment_config, reduction_config, output_config)
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()
        # run second time to reuse norm file
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()

    def test_eventfilter(self):
        self.reader_config.year = 2026
        experiment_config = options.ExperimentConfig()
        reduction_config = options.ReflectivityReductionConfig(fileIdentifier=["826"],
                                                               logfilter=['polarization_config_label==2'])
        output_config = options.ReflectivityOutputConfig(
                outputFormats=[options.OutputFomatOption.Rqz_ort],
                outputName='test',
                outputPath='test_results',
                )
        config=options.ReflectivityConfig(self.reader_config, experiment_config, reduction_config, output_config)
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()
        espin_up = reducer.dataset.data.events.shape[0]

        reduction_config.logfilter = ['polarization_config_label==3']
        output_config.append = True
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()
        espin_down = reducer.dataset.data.events.shape[0]
        # measurement should have about 2x as many counts in spin_down
        self.assertAlmostEqual(espin_down/espin_up, 2., 2)

        # perform the same filter but remove pulses during which the switch occured
        reduction_config.logfilter = ['!polarization_config_label==3']
        output_config.append = True
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()
        espin_down2 = reducer.dataset.data.events.shape[0]
        # measurement should have about 2x as many counts in spin_down
        self.assertLess(espin_down2, espin_down)

    def test_polsplitting(self):
        self.reader_config.year = 2026
        experiment_config = options.ExperimentConfig()
        reduction_config = options.ReflectivityReductionConfig(fileIdentifier=["826"])
        output_config = options.ReflectivityOutputConfig(
                outputFormats=[options.OutputFomatOption.Rqz_ort],
                outputName='test',
                outputPath='test_results',
                )
        config=options.ReflectivityConfig(self.reader_config, experiment_config, reduction_config, output_config)
        reducer = reduction_reflectivity.ReflectivityReduction(config)
        reducer.reduce()

        results = fileio.load_orso(os.path.join(output_config.outputPath, output_config.outputName+'.Rqz.ort'))
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].info.data_source.measurement.instrument_settings.polarization, 'po')
        self.assertEqual(results[1].info.data_source.measurement.instrument_settings.polarization, 'mo')
        espin_up = np.nansum(results[0].data[:,1])
        espin_down = np.nansum(results[1].data[:,1])
        # the total intensity should be around equal as events are doubled and monitor counts are doubled
        self.assertAlmostEqual(espin_down/espin_up, 1., 2)



"""
Helper functions used during calculations. Uses numba enhanced functions if available, otherwise numpy based
fallback is imported.
"""
import numpy as np
from .event_data_types import EventDatasetProtocol, append_fields

try:
    from .helpers_numba import merge_frames, extract_walltime, filter_project_x, calculate_derived_properties_focussing
except ImportError:
    from .helpers_fallback import merge_frames, extract_walltime, filter_project_x, calculate_derived_properties_focussing

def add_log_to_pulses(key, dataset: EventDatasetProtocol):
    """
    Add a log value for each pulse to the pulses array.
    """
    pulses = dataset.data.pulses
    log_data = dataset.data.device_logs[key]
    if log_data.time[0]>0:
        logTimeS = np.hstack([[0], log_data.time, [pulses.time[-1]+1]])
        logValues = np.hstack([[log_data.value[0]], log_data.value])
    else:
        logTimeS = np.hstack([log_data.time, [pulses.time[-1]+1]])
        logValues = log_data.value
    pulseLogS = np.zeros(pulses.time.shape[0], dtype=log_data.value.dtype)
    j = 0
    for i, ti in enumerate(pulses.time):
        # find the last current item that was before this pulse
        while ti>=logTimeS[j+1]:
            j += 1
        pulseLogS[i] = logValues[j]
    pulses = append_fields(pulses, [(key, pulseLogS.dtype)])
    pulses[key] = pulseLogS
    dataset.data.pulses = pulses

